--  Hey emacs!  This is -*- lua -*- code
--
--
--  Destruction for Dummies
--  An addon by Earendil of Elune
--
--
--  If you are planning on adding support modules for this (and want me
--  to incorporate them into the general release) please read and follow
--  these guidelines:
--
--  1.  Do not make any changes in my files.  I have written a generic
--      framework that should contain no class- or build- specific
--      information.  If you think you need a new general-purpose function,
--      please talk to me before writing one.  If you need one to just
--      handle something for your class or build, it should go in your file.
--
--  2.  Your code should be lightweight.  The combat routines in particular
--      are going to be called fairly frequently in a situation where the
--      user is not going to want to see a lot of lag.  Make the decision
--      rapidly and get out.  Turn on DestroDummiesPCdb.profile to find out
--      how much time your addon is taking to figure out what to do in
--      combat.  You should not need to exceed 0.07%.
--
--  3.  Do not require/permit stuff to be configured.  The name of this mod
--      was taken with some thought--the goal is to implement a "Dummies" type
--      addon, not one which allows the expert user to configure everything
--      the way they want.  You are the expert--implement the optimal rotation
--      (or at least a rotation as close to optimal as you can get).  The user
--      does not have to take your advice, but do not give choices.
--
--  4.  Do not recommend the use of long cooldown (like
--      once-per-boss-fight) casts.  Heroism for shamans, Berserk for
--      druids, and the like.  If such a cast is available, call
--      DestroDummies:altRecommend() on it.  (The problem is that in very
--      many boss fights you are going to want to save them for a specific
--      time (which the Raid Leader will probably call).  If you come up
--      with a rotation that will recommend them, the user will have no
--      way to see what you recommend if he cannot cast that yet.)
--
--  5.  Most important of all: get your rotation right.  What you
--      implement had better agree with Icy Veins take on optimal
--      rotation.  This mod will offer proven good advice, not just
--      the rotation that you happen to like.  You do not need to
--      optimize based on gear (it is going to go out of date too
--      quickly) but pick a reasonable rotation.

local altFrame = {};
altFrame.nelem = 4;

local rangeSpells = {};
rangeSpells.count = 0;

local playerClass;		-- *not* class+build; eg. "MAGE" or "WARLOCK"
local curCasting = nil;		-- spell we are currently casting

#define DEBUG(...)            		\
    if (DestroDummies.debug) then 	\
	print(__VA_ARGS__);        	\
    end



-- Return true if the passed-in name is that of a talent that we have
-- not taken.  (Note that the semantics are slightly different than
-- simply !DestroDummies:utilTalent()--if the thing is not a talent,
-- we must return false, but it is an error for utilTalent() to get
-- called with a non-talent.  For us, it is a common and expected
-- occurence.
local function is_untaken_talent (target)
    local spec_no = GetActiveSpecGroup();

    for t = 1,DD_NUM_TALENT_TIERS do
        for i = 1,DD_NUM_TALENTS_PER_TIER do
	    -- we seem to return true or false twice?  I wish this
	    -- were documented
            local _, name, _, picked1, picked2 = GetTalentInfo(t,i,spec_no);
            if (name == target) then
	        return not picked1;
	    end
	end
    end
    return false;
end

-- Figure out whether or not we know the spell.
local function is_known_spell (spell)
    -- There are some spells that GetSpellInfo() returns are unknown
    -- when called by name, even if they are known.  Like Doom.  I
    -- think it is everything with a second returned variable (what
    -- used to be Rank).  In any case, we need to handle those
    -- separately.  Fucking Bliz.
    if spell == 603 then	-- doom
	return true;		-- what the hell, I cannot think of better
    end

    -- Find the name by spell number.  This will succeed whether we
    -- know the spell or not
    local sname = GetSpellInfo(spell);

    -- Now find the name by spell name.  This will succeed only if we
    -- know the spell. (It will, however, fail in some cases where it
    -- should succeed.  Such as with Doom.  Fucking Bliz.)
    sname = GetSpellInfo(sname);

    return sname ~= nil;
end


function DestroDummies.events.ADDON_LOADED(addon)
    if addon ~= "DestructionForDummies" then
	return
    end

    print("Destruction for Dummies (v", GetAddOnMetadata("DestructionForDummies", "Version"), ") (configure with /dfd)");

    DestroDummies:LoadSavedVars();

    DestroDummies.updateInterval = DestroDummiesPCdb.noCombatUpdateInterval;
    DestroDummies.timeSinceLastUpdate = 0;
    DestroDummies.lastAltWarnTime = 0;
    DestroDummies.execTime = 0;

    -- Register for Slash Commands
    DestroDummies:CreateOptionFrame();
    SlashCmdList["DestroDummies"] = DestroDummies.Options
    SLASH_DestroDummies1 = "/DestroForDummies"
    SLASH_DestroDummies2 = "/dfd"

    SlashCmdList["DestroDummiesEnable"] = DestroDummies.CmdEnabled;
    SLASH_DestroDummiesEnable1 = "/DestroForDummies enable"
    SLASH_DestroDummiesEnable2 = "/DestroForDummies e"
    SLASH_DestroDummiesEnable3 = "/DestroForDummies en"
    SLASH_DestroDummiesEnable4 = "/dfd enable"
    SLASH_DestroDummiesEnable5 = "/dfd e"
    SLASH_DestroDummiesEnable6 = "/dfd en"

    DestroDummies:CreateGUI();

    if DestroDummiesPCdb.enabled then
	DestroDummies.turnOn();
    end

    DestroDummies:initRangeCheckSpells();
end

-- Begin running Destruction for Dummies
function DestroDummies:turnOn()
    DestroDummies.eventFrame:SetScript("OnUpdate", function(this, elapsed)
	DestroDummies:OnUpdate(elapsed)
    end)

    DestroDummies.eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
    DestroDummies.eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
    DestroDummies.eventFrame:RegisterEvent("PLAYER_REGEN_DISABLED")
    DestroDummies.eventFrame:RegisterEvent("PLAYER_TALENT_UPDATE")
    DestroDummies.eventFrame:RegisterEvent("UNIT_SPELLCAST_SENT")
    DestroDummies.eventFrame:RegisterEvent("UNIT_SPELLCAST_START")
    DestroDummies.eventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
    DestroDummies.eventFrame:RegisterEvent("UNIT_SPELLCAST_FAILED")
    DestroDummies.eventFrame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
    DestroDummies.eventFrame:RegisterEvent("UNIT_COMBAT")
end

-- Stop running Destruction for Dummies
function DestroDummies:turnOff()
    DestroDummies.eventFrame:SetScript("OnUpdate", nil);
    DestroDummies.eventFrame:UnregisterAllEvents();
    DestroDummies.suggestFrame:Hide();
end

function DestroDummies:LoadSavedVars()
    if not DestroDummiesPCdb then
	DestroDummiesPCdb = {} -- fresh start
    end

    if DestroDummiesPCdb.locked == nil then
	DestroDummiesPCdb.locked = false
    end
    if DestroDummiesPCdb.dummyWhenUnlocked == nil then
	DestroDummiesPCdb.dummyWhenUnlocked = false
    end

    if not DestroDummiesPCdb.combatUpdateInterval then
	DestroDummiesPCdb.combatUpdateInterval = 0.5
	DestroDummiesPCdb.noCombatUpdateInterval = 0.5
    end
    if DestroDummiesPCdb.enabled == nil then
	DestroDummiesPCdb.enabled = true;
    end
    if not DestroDummiesPCdb.scale then
	DestroDummiesPCdb.scale = 0.70
    end
    if not DestroDummiesPCdb.suggestPoint then
	DestroDummiesPCdb.suggestPoint = "CENTER";
	DestroDummiesPCdb.suggestXoff = 0;
	DestroDummiesPCdb.suggestYoff = 0;
	DestroDummiesPCdb.suggestRelativeTo = nil;
	DestroDummiesPCdb.suggestRelativePoint = "CENTER";
    end
    if DestroDummiesPCdb.ferralMode == nil then
	DestroDummiesPCdb.ferralMode = false
    end
end

-- This is the workhorse routine.  It gets called on every update tick
function DestroDummies:OnUpdate(elapsed)
    -- if we are forcing an update, just clear the forced update flag...
    if (DestroDummies.forceUpdate == 1) then
	DestroDummies.forceUpdate = 0;
    else
	-- otherwise, check to see if it is time to do anything yet
	self.timeSinceLastUpdate = self.timeSinceLastUpdate + elapsed;
	if (self.timeSinceLastUpdate < DestroDummies.updateInterval) then
	    -- nope, not yet time to work
	    return
	end
    end
    self.timeSinceLastUpdate = 0

    -- Test for enabled after handling the update to
    -- timeSinceLastUpdate, though as I think about it, it ought to be
    -- safe enough to ignore that.  Especially if we light forceUpdate
    -- on being enabled.
    if not DestroDummiesPCdb.enabled then
	return
    end

    -- Because talent information is not available at addon load time
    -- (thank you so very much, Blizzard), we need to burn cycles here
    -- to check and init ourselves.  We are going to do a sin of
    -- commission and trust that checking the combat routine is
    -- sufficient--that we will never have it set and the noCombat
    -- routine set to nil (guaranteed safe, as long as the rotation
    -- files are generated from dddl template files, which they always
    -- will be).  This ought to be safe, so it is worth it.  But it is
    -- still totally fucking stupid that I have to burn time every
    -- update doing the fucking stupid test and I hate the idea of
    -- having to test two things.
    --
    if (DestroDummies.decisionCombat == nil) then
    	DestroDummies:populateRegistry();
	if (DestroDummies.decisionCombat == nil) then
	    return;
	end
    end

    local spell = nil;		-- default
    local cooldown, is_item;

    local i;
    for i = 1,altFrame.nelem do
    	altFrame[i].spell = nil
    end

    -- do not add insult to injury by recommending anything when they are dead
    -- also, do not make a suggestion when they are in a vehicle.  they will
    -- have other abilities we do not know about
    if ((not UnitIsDeadOrGhost("player")) and
        (not UnitInVehicle("player")) and
        (not UnitOnTaxi("player")) and
	(not C_PetBattles.IsInBattle())) then
	-- force a reload of target debuffs
	DestroDummies.needNewDebuffData = 1;

	if (UnitAffectingCombat("player")) then
	    if (UnitExists("target") and UnitCanAttack("player", "target") and
			not UnitIsDead("target")) then
		local startTime = GetTime();
		spell, cooldown, is_item = DestroDummies.decisionCombat();
		DestroDummies.execTime = DestroDummies.execTime +
							GetTime() - startTime;
	    end
	else
	    spell, cooldown, is_item = DestroDummies.decisionNoCombat();
	end
    end

    if (DestroDummies.debug) then
	print("Decided on ", DestroDummies.spell_map[spell], " (#", spell, ")  cooldown: ", cooldown, "  is_item: ", is_item);
	DestroDummies.debug = false
    end

    if (spell == nil) then
	DestroDummies:displaySpellInFrame(DestroDummies.suggestFrame,
					  DestroDummies.suggestTexture,
					  nil, nil);
    else
	DestroDummies:displaySpellInFrame(DestroDummies.suggestFrame,
					  DestroDummies.suggestTexture,
					  spell, is_item);
    end

    local i;
    for i = 1,altFrame.nelem do
	DestroDummies:displaySpellInFrame(altFrame[i].frame,
					  altFrame[i].texture,
					  altFrame[i].spell,
					  nil); -- alt spells cannot be items
    end
end

-- Display the passed spell's texture in the passed spell's frame
function DestroDummies:displaySpellInFrame(frame, texture, spell, is_item)
    if (spell == nil) then
	if (not DestroDummies.inCombat and not DestroDummiesPCdb.locked and
			    DestroDummiesPCdb.dummyWhenUnlocked) then
	    texture:SetTexture(GetItemIcon("item:42992"));
	    frame:Show();
	else
	    texture:SetTexture(nil);
	    frame:Show();
	    --texture:SetTexture(GetItemIcon("item:42992"));
	    --frame:Hide();
	end
    else
	if is_item then
	   texture:SetTexture(GetItemIcon(spell));
	else
	   -- We must use the spell name, not id number
	   local name = DestroDummies.spell_map[spell];
	   local tex = GetSpellTexture(name);
	   if (tex == nil) then
	       -- There are some names that do not seem to work (like
	       -- "Doom" for Demonology warlocks).  In those cases, we
	       -- need to find the spell in the spellbook
	       local id = DestroDummies:munge(name, "spell", "spellID");
	       if (id ~= 0) then
		   tex = GetSpellTexture(id, BOOKTYPE_SPELL);
	       end
	   end
	   if (tex == nil) then
	       DEBUG("Cannot find texture for ", name);
	       tex = GetItemIcon("item:42992");	-- a ? icon would be better
	   end
	   texture:SetTexture(tex);
	end
	frame:Show();
    end
end


-- We have started a spell cast.  Show cooldown animation that will
-- expire when we are next able to cast a spell; this will be the max
-- of the GCD (if this spell triggers it) and the cast time of this
-- spell
function DestroDummies:showCooldownAnimation ()
    -- Start by getting the information for the spell we are now casting
    local curspell, nameSubtext, text, texture, curStartTime, curEndTime, isTradeSkill, castID, notInterruptible = UnitCastingInfo("player");

    if (curspell == nil) then
	-- We have already completed the cast, so either an instant cast or
	-- we are being told about the completion event
	curStartTime = 0;
	curEndTime = 0;
    else
	-- convert to seconds
	curStartTime = curStartTime / 1000;
	curEndTime = curEndTime / 1000;
    end
    local curDuration = curEndTime - curStartTime;

    -- Check the global cooldown.
    local cdStartTime, cdDuration = DestroDummies:GetGCD();
    if cdStartTime == nil then
        cdStartTime = 0;
    end
    if cdDuration == nil then
    	cdDuration = 0;
    end

    -- figure out which is longest
    local startTime, duration;
    if (curEndTime > (cdStartTime + cdDuration)) then
	startTime = curStartTime;
	duration = curDuration;
    else
	startTime = cdStartTime;
	duration = cdDuration;
    end

    -- do not refresh the animation if it is unchanged
    if not ((startTime == DestroDummies.svCDStartTime) and
	    (duration == DestroDummies.svCDDuration)) then

        DestroDummies.svCDStartTime = startTime;
	DestroDummies.svCDDuration = duration;

	-- Display the cooldown animation for the long pole
	DestroDummies.cooldownFrame:SetCooldown(startTime, duration)
    end
end


function DestroDummies:CreateGUI()
    --
    --  Create the main suggest frame
    --
    DestroDummies.suggestFrame = CreateFrame("Frame", nil, UIParent);
    local f = DestroDummies.suggestFrame;

    local t = f:CreateTexture(nil, "BACKGROUND");
    t:SetTexture(nil)
    t:SetAllPoints(f)
    t:SetAlpha(0.5)
    f.texture = t
    DestroDummies.suggestTexture = t

    f:SetFrameStrata("BACKGROUND");
    f:SetClampedToScreen(true);
    f:SetWidth(128);
    f:SetHeight(128);
    f:SetScale(DestroDummiesPCdb.scale);
    f:SetPoint(DestroDummiesPCdb.suggestPoint, DestroDummiesPCdb.suggestRelativeFrame, DestroDummiesPCdb.suggestRelativePoint, DestroDummiesPCdb.suggestXoff, DestroDummiesPCdb.suggestYoff);
    f:Show();


    --
    --  Create the side frames
    --
    local i;
    for i = 1,altFrame.nelem do
        local part1, part2a, part2b;

    	if (i == 1 or i == 3) then
	    part1 = "BOTTOM";
	else
	    part1 = "TOP";
	end
	if (i == 1 or i == 2) then
	    part2a = "RIGHT";
	    part2b = "LEFT";
	else
	    part2a = "LEFT";
	    part2b = "RIGHT";
	end

    	altFrame[i] = {};
	altFrame[i].frame, altFrame[i].texture =
		DestroDummies:createFrame(64, part1..part2a, f, part1..part2b);
    end


    f:RegisterForDrag("LeftButton");
    f:SetScript("OnDragStart", f.StartMoving);
    f:SetScript("OnDragStop", DestroDummies.stopMoving);
    DestroDummies:FrameOptions();


    -- cooldown animation
    DestroDummies.cooldownFrame = CreateFrame("Cooldown", "myCooldown",
					      f, "CooldownFrameTemplate");
    DestroDummies.cooldownFrame:SetAllPoints();
    DestroDummies.cooldownFrame:SetDrawEdge(false);
    DestroDummies.cooldownFrame:SetDrawSwipe(true);  -- do the rotating wipe
end

function DestroDummies:stopMoving()
    local f = DestroDummies.suggestFrame;

    f:StopMovingOrSizing();
    local point, relativeTo, relativePoint, xOff, yOff = f:GetPoint();
    DestroDummiesPCdb.suggestPoint = point;
    DestroDummiesPCdb.suggestRelativeTo = relativeTo;
    DestroDummiesPCdb.suggestRelativePoint = relativePoint;
    DestroDummiesPCdb.suggestXoff = xOff;
    DestroDummiesPCdb.suggestYoff = yOff;
end

function DestroDummies:createFrame(size, myPoint, parent, parentPoint)
    local f = CreateFrame("Frame", nil, parent);
    local t = f:CreateTexture(nil, "BACKGROUND");
    t:SetTexture(nil);
    t:SetAllPoints(f);
    t:SetAlpha(0.5);
    f.texture = t;

    f:SetFrameStrata("BACKGROUND");
    f:SetClampedToScreen(true);
    f:SetWidth(size);
    f:SetHeight(size);
    f:SetPoint(myPoint, parent, parentPoint);

    return f,t;
end

-- Set various frame options for combat/no combat
function DestroDummies:FrameOptions()
    local f = DestroDummies.suggestFrame;
    if (DestroDummies.inCombat or DestroDummiesPCdb.locked) then
        f:SetMovable(false);
        f:EnableMouse(false);
    else
        f:SetMovable(true);
        f:EnableMouse(true);
    end
end

-- called for every attack or heal
function DestroDummies.events.UNIT_COMBAT(unit, action, descriptor, damage,
					  damageType)
    if (unit ~= "player") then
        return;
    end

    if action == "WOUND" then
       if DestroDummies.takeDamage ~= nil then
	  DestroDummies.takeDamage(damage);
       end
    elseif action == "HEAL" then
       if DestroDummies.takeHeal ~= nil then
	  DestroDummies.takeHeal(damage);
       end
    end
end

function DestroDummies.events.UNIT_SPELLCAST_SENT(unit, loc_spell)
    local spell;

    -- the word is "player" even in spanish, and I suspect all other languages
    if (unit ~= "player") then
        return;
    end

    -- translate the spell into English
    spell = GetSpellInfo(loc_spell);

    if DestroDummies.trace_events then
	print("UNIT_SPELLCAST_SENT(", unit, spell, ");");
    end

    -- we used to just not overwrite lastSpell on auto shot, but did
    -- everything else, which seems wrong to me.
    if (spell == "Auto Shot") then
	return;
    end

    curCasting = spell;
    DestroDummies.forceUpdate = 1;
    DestroDummies.lastSpell = spell;
    DestroDummies.castStartTime = GetTime();

    if (DestroDummies.spellSent ~= nil) then
    	DestroDummies.spellSent(spell);
    end
end

function DestroDummies.events.UNIT_SPELLCAST_START(unit, spell)
    if (unit ~= "player") then
        return;
    end

    if DestroDummies.trace_events then
	print("UNIT_SPELLCAST_START(", unit, spell, ");");
    end

    DestroDummies:showCooldownAnimation();
end

function DestroDummies.events.UNIT_SPELLCAST_SUCCEEDED(unit, spell)
    if (unit ~= "player") then
        return;
    end

    if DestroDummies.trace_events then
	print("UNIT_SPELLCAST_SUCCEEDED(", unit, spell, ");");
    end

    curCasting = nil;
    DestroDummies.forceUpdate = 1;

    DestroDummies:showCooldownAnimation();

    -- UNIT_SPELLCAST_START is not called on instant casts, so need to
    -- set here too
    if (spell ~= "Auto Shot") then	-- Do not let auto shot overwrite
       DestroDummies.lastSpell = spell;
    end

    if (DestroDummies.postSpell ~= nil) then
    	DestroDummies.postSpell(spell, true);
    end
end

function DestroDummies.events.UNIT_SPELLCAST_FAILED(unit, spell)
    if (unit ~= "player") then
        return;
    end

    if DestroDummies.trace_events then
	print("UNIT_SPELLCAST_FAILED(", unit, spell, ");");
    end

    curCasting = nil;
    DestroDummies.forceUpdate = 1;
    DestroDummies.lastSpell = nil;

    DestroDummies:showCooldownAnimation();

    if (DestroDummies.postSpell ~= nil) then
    	DestroDummies.postSpell(spell, false);
    end
end

function DestroDummies.events.UNIT_SPELLCAST_INTERRUPTED(unit, spell)
    if (unit ~= "player") then
        return;
    end

    if DestroDummies.trace_events then
	print("UNIT_SPELLCAST_INTERRUPTED(", unit, spell, ");");
    end

    DestroDummies:showCooldownAnimation();
end

function DestroDummies.events.PLAYER_TALENT_UPDATE()
    DestroDummies:populateRegistry();
    --DestroDummies:uniquify_spells();
    DestroDummies:initRangeCheckSpells();
end

function DestroDummies.events.PLAYER_TARGET_CHANGED()
    DestroDummies.lastSpell = nil;
end

-- enter combat
function DestroDummies.events.PLAYER_REGEN_DISABLED()
    if (DestroDummies.preCombat ~= nil) then
    	DestroDummies.preCombat();
    end

    DestroDummies.combatStartTime = GetTime();
    DestroDummies.execTime = 0;

    DestroDummies.updateInterval = DestroDummiesPCdb.combatUpdateInterval;
    DestroDummies.forceUpdate = 1;
    DestroDummies:FrameOptions();

    DestroDummies.inCombat = true;
    DestroDummies:FrameOptions();
end

-- exit combat
function DestroDummies.events.PLAYER_REGEN_ENABLED()
    local endTime = GetTime();
    DestroDummies.inCombat = false;
    DestroDummies:FrameOptions();

    DestroDummies.updateInterval = DestroDummiesPCdb.noCombatUpdateInterval;
    DestroDummies.forceUpdate = 1;
    DestroDummies:FrameOptions();

    if DestroDummiesPCdb.profile then
        local combatTime = endTime - DestroDummies.combatStartTime;
        print("Destruction for Dummies: combat ended after ", combatTime,
                "seconds.");
        print("  -- DfD took ", DestroDummies.execTime, " seconds (",
                (DestroDummies.execTime * 100) / combatTime, "%");
    end
    DestroDummies.combatStartTime = 0;
end

function DestroDummies:populateRegistry()
    DestroDummies.myClass = DestroDummies:whatAmI();
    if (DestroDummies.myClass == classUnknown) then
        DestroDummies.preCombat = nil;
	DestroDummies.spellSent = nil;
	DestroDummies.postSpell = nil;
        DestroDummies.takeHeal = nil;
        DestroDummies.takeDamage = nil;
        DestroDummies.decisionCombat = nil;
        DestroDummies.decisionNoCombat = nil;
    else
        DestroDummies.preCombat = DestroDummies.registry.preCombat[DestroDummies.myClass];
        DestroDummies.spellSent = DestroDummies.registry.spellSent[DestroDummies.myClass];
        DestroDummies.postSpell = DestroDummies.registry.postSpell[DestroDummies.myClass];
        DestroDummies.takeDamage = DestroDummies.registry.takeDamage[DestroDummies.myClass];
        DestroDummies.takeHeal = DestroDummies.registry.takeHeal[DestroDummies.myClass];
        DestroDummies.decisionCombat = DestroDummies.registry.decideCombat[DestroDummies.myClass];
        DestroDummies.decisionNoCombat = DestroDummies.registry.decideNoCombat[DestroDummies.myClass];
    end
end

function DestroDummies.Options()
    InterfaceOptionsFrame_OpenToCategory(getglobal("DestroDummiesOptions"));
end

function DestroDummies.CmdEnabled()
    if (DestroDummiesPCdb.enabled == true) then
       print("Disabling Destruction for Dummies");
    else
       print("Enabling Destruction for Dummies");
    end
    DestroDummies:ToggleEnabled();
end

function DestroDummies:CreateOptionFrame()
    local panel = CreateFrame("FRAME", "DestroDummiesOptions");
    panel.name = "Destruction for Dummies";

    local fstring1 = panel:CreateFontString("DestroDummiesOptions_1_string1","OVERLAY","GameFontNormal")
    fstring1:SetText(DestroDummies:xlateString("Enable (for this Character)"));
    fstring1:SetPoint("TOPLEFT", 10, -10)
    local checkbox1 = CreateFrame("CheckButton", "$parent_cb1", panel, "OptionsCheckButtonTemplate")
    checkbox1:SetWidth(30)
    checkbox1:SetHeight(30)
    checkbox1:SetScript("OnClick", function() DestroDummies:ToggleEnabled() end)
    checkbox1:SetPoint("TOPRIGHT", -10, -10)
    checkbox1:SetChecked(DestroDummiesPCdb.enabled);


    local fstring2 = panel:CreateFontString("DestroDummiesOptions_1_string2","OVERLAY","GameFontNormal")
    fstring2:SetText(DestroDummies:xlateString("Reposition Lock"));
    fstring2:SetPoint("TOPLEFT", 10, -50)
    local checkbox2 = CreateFrame("CheckButton", "$parent_cb1", panel, "OptionsCheckButtonTemplate")
    checkbox2:SetWidth(30)
    checkbox2:SetHeight(30)
    checkbox2:SetScript("OnClick", function() DestroDummies:ToggleLocked() end)
    checkbox2:SetPoint("TOPRIGHT", -10, -50)
    checkbox2:SetChecked(DestroDummiesPCdb.locked)


    local fstring3 = panel:CreateFontString("DestroDummiesOptions_1_string3","OVERLAY","GameFontNormal")
    fstring3:SetText(DestroDummies:xlateString("Display dummy suggestions when unlocked"))
    fstring3:SetPoint("TOPLEFT", 10, -90)
    local checkbox3 = CreateFrame("CheckButton", "$parent_cb1", panel, "OptionsCheckButtonTemplate")
    checkbox3:SetWidth(30)
    checkbox3:SetHeight(30)
    checkbox3:SetScript("OnClick", function() DestroDummies:ToggleDummySuggest() end)
    checkbox3:SetPoint("TOPRIGHT", -10, -90)
    checkbox3:SetChecked(DestroDummiesPCdb.dummyWhenUnlocked);


    local fstring8 = panel:CreateFontString("DestroDummiesOptions_1_string8","OVERLAY","GameFontNormal")
    fstring8:SetText(DestroDummies:xlateString("Minimal assistance mode"))
    fstring8:SetPoint("TOPLEFT", 10, -130)
    local checkbox8 = CreateFrame("CheckButton", "$parent_cb1", panel, "OptionsCheckButtonTemplate")
    checkbox8:SetWidth(30)
    checkbox8:SetHeight(30)
    checkbox8:SetScript("OnClick", function() DestroDummies:ToggleFerralMode() end)
    checkbox8:SetPoint("TOPRIGHT", -10, -130)
    checkbox8:SetChecked(DestroDummiesPCdb.ferralMode);

    local fstring6 = panel:CreateFontString("DestroDummiesOptions_1_string6","OVERLAY","GameFontNormal")
    fstring6:SetText("Scale")
    fstring6:SetPoint("TOPLEFT", 10, -170)
    local slider2 = CreateFrame("Slider", "$parent_sl2", panel, "OptionsSliderTemplate")
    slider2:SetMinMaxValues(.25, 8.0)
    slider2:SetValue(DestroDummies:GetScale())
    slider2:SetValueStep(.25)
    slider2:SetScript("OnValueChanged", function(self) DestroDummies:SetScale(self:GetValue()); getglobal(self:GetName() .. "Text"):SetText(self:GetValue())  end)
    getglobal(slider2:GetName() .. "Low"):SetText("0.25")
    getglobal(slider2:GetName() .. "High"):SetText("8.0")
    getglobal(slider2:GetName() .. "Text"):SetText(DestroDummies:GetScale())
    slider2:SetPoint("TOPRIGHT", -10, -170)

    local resetButton = CreateFrame("Button", "$parent_rb", panel, "OptionsButtonTemplate");
    resetButton:SetText(DestroDummies:xlateString("Reset"));
    resetButton:SetPoint("TOPRIGHT", -20, -210);
    resetButton:SetScript("OnClick", DestroDummies.Reset);

    InterfaceOptions_AddCategory(panel);
end

function DestroDummies:Reset()
    DestroDummiesPCdb.scale = 0.70
    DestroDummiesPCdb.suggestPoint = "CENTER";
    DestroDummiesPCdb.suggestXoff = 0;
    DestroDummiesPCdb.suggestYoff = 0;
    DestroDummiesPCdb.suggestRelativeTo = nil;
    DestroDummiesPCdb.suggestRelativePoint = "CENTER";

    local f = DestroDummies.suggestFrame;
    f:ClearAllPoints();
    f:SetPoint(DestroDummiesPCdb.suggestPoint, DestroDummiesPCdb.suggestRelativeFrame, DestroDummiesPCdb.suggestRelativePoint, DestroDummiesPCdb.suggestXoff, DestroDummiesPCdb.suggestYoff);
    f:SetScale(DestroDummiesPCdb.scale);
    f:Show();
end

function DestroDummies:ToggleLocked()
    DestroDummiesPCdb.locked = not DestroDummiesPCdb.locked;
    DestroDummies:FrameOptions();
    DestroDummies.forceUpdate = 1;
end

function DestroDummies:ToggleFerralMode()
    DestroDummiesPCdb.ferralMode = not DestroDummiesPCdb.ferralMode;
    DestroDummies:FrameOptions();
    DestroDummies.forceUpdate = 1;
end

function DestroDummies:ToggleDummySuggest()
    DestroDummiesPCdb.dummyWhenUnlocked = not DestroDummiesPCdb.dummyWhenUnlocked;
    DestroDummies.forceUpdate = 1;
end

function DestroDummies:ToggleEnabled()
    if (DestroDummiesPCdb.enabled == true) then
        DestroDummiesPCdb.enabled = false;
	DestroDummies.turnOff();
    else
        DestroDummiesPCdb.enabled = true;
	DestroDummies.turnOn();
    end
    DestroDummies:FrameOptions();
    DestroDummies.forceUpdate = 1;
end

-- API for people to use in macros
function DestroDummies:SetEnabled()
    DestroDummiesPCdb.enabled = true;
    DestroDummies.turnOn();
    DestroDummies:FrameOptions();
    DestroDummies.forceUpdate = 1;
end

-- API for people to use in macros
function DestroDummies:SetDisabled()
    DestroDummiesPCdb.enabled = false;
    DestroDummies.turnOff();
    DestroDummies:FrameOptions();
    DestroDummies.forceUpdate = 1;
end

function DestroDummies:GetScale()
    return DestroDummiesPCdb.scale;
end

function DestroDummies:SetScale(scale)
    DestroDummiesPCdb.scale = scale;
    DestroDummies.suggestFrame:SetScale(scale);
end

function DestroDummies:checkBeta(class)
    if (DestroDummies.beta) then
        return class;
    else
        return classUnknown;
    end
end

-- A debugging function.  Set a class to force, and call the talent update
-- handler to get us to reassign everything
--
function DestroDummies:forceClass(new)
    DestroDummies.forcedClass = new;
    DestroDummies.events.PLAYER_TALENT_UPDATE();
end

-- return the index for our class, or 0 if we do not know about it
function DestroDummies:whatAmI()
    -- For debugging
    if (DestroDummies.forcedClass ~= nil and DestroDummies.forcedClass ~= 0) then
        return DestroDummies.forcedClass;
    end

    local pc1
    pc1, playerClass = UnitClass("player");

    -- maxTab will hold the talent specialization number.  (This is a holdover
    -- name from pre Panda days when we had to count the number of points in
    -- each group to figure out what specialization this was)
    local maxTab;
    maxTab = GetSpecialization();

    if playerClass == "DEATHKNIGHT" then
        if (maxTab == 1) then
	    return DestroDummies.classDeathknightBlood;
        end
        if (maxTab == 2) then
	    return DestroDummies.classDeathknightFrost;
        end
        if (maxTab == 3) then
	    return DestroDummies.classDeathknightUnholy;
        end
        return 0;
    end

    if playerClass == "DRUID" then
        if (maxTab == 1) then
            return DestroDummies.classDruidBalance;
        end
        if (maxTab == 2) then		-- Feral
	    -- Feral and guardian share the same rotation file for
	    -- historical (and practical) reasons.  So this is a hack
	    -- to let me tell them apart when needed.  (At some point,
	    -- they really should be separated into different files,
	    -- but that is a long way off)
	    DestroDummies.is_feral = true;
            return DestroDummies.classDruidFeral;
        end
        if (maxTab == 3) then		-- Guardian
	    DestroDummies.is_feral = false;
            return DestroDummies.classDruidFeral;
        end
        return 0;
    end

    if playerClass == "HUNTER" then
        if (maxTab == 1) then
            return DestroDummies.classHunterBeastmaster;
        end
        if (maxTab == 2) then
            return DestroDummies.classHunterMarksman;
        end
        if (maxTab == 3) then
            return DestroDummies.classHunterSurvival;
        end
        return 0;
    end

    if playerClass == "MAGE" then
        if (maxTab == 1) then
            return DestroDummies.classMageArcane;
        end
        if (maxTab == 2) then
            return DestroDummies.classMageFire;
        end
        if (maxTab == 3) then
            return DestroDummies.classMageFrost;
        end
        return 0;
    end

    if playerClass == "PALADIN" then
        if (maxTab == 3) then
            return DestroDummies.classPaladinRetribution;
        end
	return 0;
    end

    if playerClass == "PRIEST" then
        if (maxTab == 3) then
            return DestroDummies.classPriestShadow;
        end
	return 0;
    end

    if playerClass == "ROGUE" then
        if (maxTab == 1) then
	    return DestroDummies.classRogueAssassination;
	end
        if (maxTab == 2) then
	    return DestroDummies.classRogueCombat;
	end
        if (maxTab == 3) then
	    return DestroDummies.classRogueSubtlety;
	end
    end

    if playerClass == "SHAMAN" then
        if (maxTab == 1) then
            return DestroDummies.classShamanElemental;
        end
        if (maxTab == 2) then
            return DestroDummies.classShamanEnhancement;
        end
	return 0;
    end

    if playerClass == "WARLOCK" then
        if (maxTab == 1) then
            return DestroDummies.classWarlockAffliction;
        end
        if (maxTab == 2) then
            return DestroDummies.classWarlockDemonology;
        end
        if (maxTab == 3) then
            return DestroDummies.classWarlockDestruction;
        end
        return 0;
    end

    if playerClass == "WARRIOR" then
        if (maxTab == 2) then
            return DestroDummies.classWarriorFury;
        end
	if (maxTab == 3) then
            return DestroDummies.classWarriorProtection;
        end
	return 0;
    end

    if playerClass == "MONK" then
        if (maxTab == 3) then
            return DestroDummies.classMonkWindwalker;
        end
	return 0;
    end

    return 0;
end


-- return true if player knows the spell (of any rank), false otherwise
function DestroDummies:knowsSpell(target)
    local i;
    for i = 1, MAX_SKILLLINE_TABS do
        local name, _, offset, numSpells = GetSpellTabInfo(i);

        if not name then
            break
        end

        for s = offset + 1, offset + numSpells do
            local spell, rank = GetSpellBookItemName(s, BOOKTYPE_SPELL);
            if spell == target then
                return true;
            end
        end
    end
    return false;
end


-- return true if we can cast the passed spell
function DestroDummies:spellReady (spell)
    local castable, mana = IsUsableSpell(spell);
    if (not castable or mana) then
    	return false;
    end

    local now = GetTime();

    -- compute the cooldown specific to this spell
    local start, duration, enabled = GetSpellCooldown(spell);
    local spell_cooldown = start + duration - now;
    if (spell_cooldown < 0) then
	spell_cooldown = 0;
    end

    -- it is ready if it is cooldown is less than or equal to the GCD.
    local gcd_start, gcd_dur = DestroDummies:GetGCD();
    local gcd_cooldown = gcd_start + gcd_dur - now;
    if (gcd_cooldown < 0) then
	gcd_cooldown = 0;
    end

    local result = spell_cooldown <= gcd_cooldown;
    if DestroDummies.debug then
	print("spellReady(", DestroDummies:munge(spell, "spell", "string"),
	      "(#", spell, ")): ", spell_cooldown, ", ",
	      gcd_cooldown, " = ", result);
    end
    return result;
end


-- refresh our knowledge of what debuffs are up on target
function DestroDummies:refreshDebuffData ()
    DestroDummies.needNewDebuffData = false;
    DestroDummies.numDebuffsByAny = 0;
    DestroDummies.numDebuffsByMe = 0;
    for i = 1,40 do
        local name,_,_,count,_,_,expiry,caster,_,_,spid = UnitDebuff("target", i);

        if (name ~= nil) then
	    local j = DestroDummies.numDebuffsByAny;
	    DestroDummies.numDebuffsByAny = j + 1;
	    j = DestroDummies.numDebuffsByAny;

	    DestroDummies.debuffByAny[j].name = name;
	    DestroDummies.debuffByAny[j].expiry = expiry;
	    DestroDummies.debuffByAny[j].count = count;
	    DestroDummies.debuffByAny[j].id = spid;
	    DestroDummies.debuffByAny[j].caster = caster;

	    if (caster == "player") then
		j = DestroDummies.numDebuffsByMe;
		DestroDummies.numDebuffsByMe = j + 1;
		j = DestroDummies.numDebuffsByMe;

		DestroDummies.debuffByMe[j].name = name;
		DestroDummies.debuffByMe[j].expiry = expiry;
		DestroDummies.debuffByMe[j].count = count;
		DestroDummies.debuffByMe[j].id = spid;
		DestroDummies.debuffByMe[j].caster = caster;
	    end
	end
    end
end

-- return the amount of time left on the buff (or nil if it is not up), along
-- with the count of how many are up.
--
-- Note well that a buff that does not expire will return 0, so a test like
-- if (DestroDummies:buffTime("Aspect of the Hawk")) will never return true.
-- the only possible return values for Aspect of the Hawk (which does not
-- expire) are nil and 0.
--
function DestroDummies:buffTime(buff, cretin)
    if cretin == nil then
       cretin = "player";
    end

    -- The UnitAura only works with names, not indexes, so ensure that buff
    -- is the name
    buff = DestroDummies:aura_name(buff);

    local name,_,_,count,_,_, expiry = UnitAura(cretin, buff, nil, "HELPFUL");
    if (name == nil) then
        return nil, 0;
    else
        return expiry - GetTime(), count;
    end
end

function DestroDummies:debuffTime(debuff, cretin)
    if cretin == nil then
       cretin = "player";
    end

    -- The UnitAura only works with names, not indexes, so ensure that buff
    -- is the name
    buff = DestroDummies:aura_name(buff);

    local name,_,_,count,_,_, expiry = UnitAura(cretin, debuff,nil, "HARMFUL");
    if (name == nil) then
	return nil, 0;
    else
	return expiry - GetTime(), count;
    end
end

-- A debugging function, to display all the talents, along with their
-- ids.  Useful when creating new rotations
function DestroDummies:showTalents ()
    local spec_no = GetActiveSpecGroup();

    for t = 1,DD_NUM_TALENT_TIERS do
        for i = 1,DD_NUM_TALENTS_PER_TIER do
            local id, name, icon, picked1,picked2 = GetTalentInfo(t,i,spec_no);
	    print("Talent ",t,".",i,": ", name, "(",id,")");
	end
    end
    return false;
end

-- Display the spell id for the passed-in spell
--
-- Damned useful when creating a rotation.  Note that in some cases,
-- this yields an incorrect number (eg., "Glaive Toss" and "Barrage"
-- both yield 175687, which is actually "Heavy Artillery"); the actual
-- numbers for those spells differ, and can be found by calling
-- GetSpellInfo("Glaive Toss") from a character that knows that spell.
--
--function ss (target) -- +++
function DestroDummies:show_spell (target)
   local tab, tabName, tabSpellOffset, tabNumSpells, spellIndex;

   for tab = 1, MAX_SKILLLINE_TABS do
      tabName, _, tabSpellOffset, tabNumSpells = GetSpellTabInfo(tab);
      if not tabName then
	 break
      end

      for spellIndex = tabSpellOffset + 1, tabSpellOffset + tabNumSpells do
	 local name, subnm = GetSpellBookItemName(spellIndex, BOOKTYPE_SPELL);
	 local sType, id = GetSpellBookItemInfo(spellIndex, BOOKTYPE_SPELL);

	 if name == target then
	    print(tabName..":"..tab.." "..name.." "..id);
	    print("Spell Info: ", GetSpellInfo(name));
	    return;
	 end
      end
   end

   print("Unknown spell: "..target);
end

-- Display all the spells and spell ids we know
--
-- If formatted is non-nil, then output will be cut-and-pastable into
-- the knownSpells[] array.  Use the /chatlog command to make the
-- chat window get written to a file and then call this function.
--
-- Sigh.  The formatted function is not working.  I cannot figure
-- out how to get the damned text window to display
function DestroDummies:showMySpells(start)
    -- used to be parameters, might be again
    local finish = start;
    local formatted = nil;

   local tab, tabName, tabSpellOffset, tabNumSpells, spellIndex, frame;

   formatted = nil;		-- for now

   if start == nil or start == 0 then
      start = 2;
      finish = 2;
   end
   if finish == nil or finish == 0 then
      finish = MAX_SKILLLINE_TABS
   end

   -- we will save this frame off in a global and reuse it
   if formatted and not DestroDummies.my_spells_frame then
      frame = CreateFrame("ScrollingMessageFrame", nil, UIParent);
      --local t = frame:CreateTexture(nil, "BACKGROUND");
      --t:SetTexture(nil);
      --t:SetTexture(GetItemIcon("item:42992"));
      --t:SetAllPoints(frame);
      --t:SetAlpha(0.5);
      --frame.texture = t;
      frame:SetFrameStrata("BACKGROUND");
      frame:SetClampedToScreen(true);
      frame:SetWidth(700);
      frame:SetHeight(500);
      frame:SetScale(1);
      frame:SetPoint("BOTTOMLEFT", 0, 0);
      frame:SetAlpha(1.0);
      frame:EnableMouse(true);
      DestroDummies.my_spells_frame = frame;
   end

   for tab = start, finish do
      tabName, _, tabSpellOffset, tabNumSpells = GetSpellTabInfo(tab);
      if not tabName then
	 break
      end

      for spellIndex = tabSpellOffset + 1, tabSpellOffset + tabNumSpells do
	 local name, subnm = GetSpellBookItemName(spellIndex, BOOKTYPE_SPELL);
	 local sType, id = GetSpellBookItemInfo(spellIndex, BOOKTYPE_SPELL);
	 if formatted then
	    DestroDummies.my_spells_frame:AddMessage("    { \""..name.."\",            "..id.." },");
	 else
	    print(tabName..":"..tab.." "..name.." "..id);
	 end
      end
   end
   if formatted then
      DestroDummies.my_spells_frame:Show();
   end
end


-- return the spell name, id, expiration time, and count for my cast
-- of the debuff
function DestroDummies:debuffUpByMe(debuff)
    if (DestroDummies.needNewDebuffData) then
        DestroDummies:refreshDebuffData();
    end

    for i = 1,DestroDummies.numDebuffsByMe do
        if (DestroDummies.debuffByMe[i].id == debuff or
	    DestroDummies.debuffByMe[i].name == debuff) then
            return DestroDummies.debuffByMe[i].name, DestroDummies.debuffByMe[i].id, DestroDummies.debuffByMe[i].expiry, DestroDummies.debuffByMe[i].count;
        end
    end
    return nil, nil, nil;
end

-- A debugging function
function DestroDummies:showBuffs(unit)
    local now = GetTime();

    if unit == nil then
    	unit = "player";
    end
    local i;

    print("Showing buffs for ", unit);
    for i = 1,40 do
	local name, _,_, count, _, _, expiry, _, _, _, id = UnitAura(unit, i);
	if (name ~= nil) then
	   print(name, " (#", id, ")  expiry: ", expiry - now, ", count: ", count);
	end
    end
end

-- A debugging function
function DestroDummies:showDebuffs(cretin, force)
    local now = GetTime();

    if cretin == "player" or cretin == nil then
	local i;
	for i = 1,40 do
	    local name, _,_, _, _, _,expiry,_,_,_,id = UnitDebuff(cretin, i);
	    if (name ~= nil) then
		print(name, "(#", id, "): ", expiry - now);
	    end
	end
    else
	if (force == 1 or DestroDummies.needNewDebuffData) then
	    DestroDummies:refreshDebuffData();
	end
	local i;
	for i = 1,DestroDummies.numDebuffsByAny do
	    print (DestroDummies.debuffByAny[i].name,
		   " (#", DestroDummies.debuffByAny[i].id,
		   "): ", DestroDummies.debuffByAny[i].expiry - now,
		   "[", DestroDummies.debuffByAny[i].caster, "]",
		   "; count: ", DestroDummies.debuffByAny[i].count);
	end
    end
end

-- return the spell name, id, expiration time, and count for any cast
-- of the debuff if the same debuff is up from multiple people, return
-- the most distant expiry
function DestroDummies:debuffUpByAny(debuff)
    if (DestroDummies.needNewDebuffData) then
        DestroDummies:refreshDebuffData();
    end

    local i, sv_entry;
    sv_entry = 0;
    for i = 1,DestroDummies.numDebuffsByAny do
        if (DestroDummies.debuffByAny[i].id == debuff or
	    DestroDummies.debuffByAny[i].name == debuff) then
	    -- if we never expire, or if this is the first we have seen, or if
	    -- this entry expires after the previous one, save off this entry
	    if ((DestroDummies.debuffByAny[i].expiry == nil) or
	        (sv_entry == 0) or
                (DestroDummies.debuffByAny[sv_entry].expiry <
                                DestroDummies.debuffByAny[i].expiry)) then
                sv_entry = i;
            end
        end
    end
    if (sv_entry == 0) then
        return nil, 0, nil, 0;
    else
        return DestroDummies.debuffByAny[sv_entry].name, DestroDummies.debuffByAny[sv_entry].id, DestroDummies.debuffByAny[sv_entry].expiry, DestroDummies.debuffByAny[sv_entry].count;
    end
end

-- recommend an alternate spell
local altTestOrder = {};
altTestOrder[1] = 1;
altTestOrder[2] = 3;
altTestOrder[3] = 2;
altTestOrder[4] = 4;
function DestroDummies:altRecommend(spell, num)
    if not is_known_spell(spell) then
	return;
    end

    if (num >= 1 and num <= altFrame.nelem) then
	if (altFrame[num].spell == nil) then
	    altFrame[num].spell = spell;
	    return
	end
    end

    //--You can't always get what you want...
    local i;
    for i = 1,altFrame.nelem do
	if (altFrame[altTestOrder[i]].spell == nil) then
	    -- But if you try, sometimes, you get what you need
	    altFrame[altTestOrder[i]].spell = spell;
	    return;
	end
    end
end

-- reset our analysis engine
local analTable = {};
local analCount = 0;
analTable[0] = {};
analTable[0].spell = nil;
analTable[0].cooldown = 0;

function DestroDummies:analysisInit()
    analCount = 0;
end


-- Add a spell to analyze.  returns the cooldown for debugging purposes only.
-- The spell must be the spell id, not the name
--
-- if dependSpell is non-nil, the cooldown cannot be earlier than the cooldown
-- for the dependSpell.  If the dependSpell is not castable then do not add
-- the spell at all
function DestroDummies:analysisAdd (spell, dependSpell, filename, lineno)
    local sname = DestroDummies:munge(spell, "spell", "string");
    local cooldown = DestroDummies:cooldownTime(spell);

    if DestroDummies.debug then
	print("Begin analysisAdd([", filename, ":", lineno, "]",
	      DestroDummies.spell_map[spell], " (#",spell,"), ",
	      dependSpell, "), cooldown ", cooldown);
    end

    -- IsUsableSpell() will return that a spell is usable even if it
    -- is not know, if it is called by number instead of by name (and
    -- calling by name breaks internationalization).  Fucking Bliz.
    if not is_known_spell(spell) then
	if DestroDummies.debug then
	    print("Spell ", sname, "(#", spell, ") is not known...skipping");
	end
	return nil;
    end

    -- see if the spell is usable
    local usable, nomana = IsUsableSpell(spell);
    if (not usable and not nomana) then
	if (DestroDummies.debug) then
	    print("Unknown spell ", sname, " (", spell, ")...skipping");
	end
	return nil;
    end

    if dependSpell then
       if not DestroDummies:spellReady(dependSpell) then
	   if DestroDummies.debug then
	      print("The dependSpell is not ready--return nil");
	   end
	   return nil;
	end

	local depCool = DestroDummies:cooldownTime(dependSpell);
	if (depCool > cooldown) then
	   cooldown = depCool;
	end
    end

    analCount = analCount + 1;
    analTable[analCount] = {};
    analTable[analCount].spell = spell;
    analTable[analCount].cooldown = cooldown;
    analTable[analCount].killed = false;
    analTable[analCount].is_item = false;
    analTable[analCount].name = sname;
    if (DestroDummies.debug) then
	print("Analysis Add spell: ", sname, " (", spell, "), cooldown: ", cooldown);
    end
    return cooldown;
end


-- add an item to analyze.  returns the cooldown for debugging purposes only
--
function DestroDummies:analysisAddItem (item, off_cooldown_only,
					filename, lineno)
    if DestroDummies.debug then
	print("Begin analysisAddItem([", filename, ":", lineno, "],",
	      item, ")");
    end

    -- figure out the cooldown
    local cstart, cduration, cenable = GetItemCooldown(item);
    local cend = cstart + cduration;
    local cooldown = cend - GetTime();
    if (cooldown < 0) then
	cooldown = 0;
    end

    if off_cooldown_only and cooldown > 0 then
	if DestroDummies.debug then
	    print("  Item is on cooldown (", cooldown, " seconds). Skipping.");
	end
	return cooldown;
    end

    analCount = analCount + 1;
    analTable[analCount] = {};
    analTable[analCount].spell = item;
    analTable[analCount].cooldown = cooldown;
    analTable[analCount].killed = false;
    analTable[analCount].is_item = true;
    analTable[analCount].name = DestroDummies.item_map[item];
    if DestroDummies.debug then
       print("analysisAddItem(", item, "): ", cooldown);
    end
    return cooldown;
end

-- pick the spell to cast.  return the spell and how long it will be before
-- you can cast it and whether it is a spell or an item
function DestroDummies:analysisPick (powerAlt, freeSpell)
    local i;
    local min;

    -- I sometimes pass 0 instead of nil for no spell.  I could fix this,
    -- but it would likely break again.  it is easier to just correct it here.
    if powerAlt == 0 then
	powerAlt = nil;
    end
    if freeSpell == 0 then
	freeSpell = nil;
    end

    -- find the shortest cooldown and return it.  in case of ties, favor
    -- the one with the lowest index.  If we cannot cast the spell in
    -- question because of shortage of power, return the powerAlt spell
    -- if they gave one.  If they cannot cast it because they do not know
    -- it, kill it in the table and find the next best
    while true do
	-- pick the best of the remaining spells
	min = 0;
        for i = 1,analCount do
	    if (not analTable[i].killed) then
	    	if (min == 0) then
		    min = i;
	    	elseif ((analTable[i].cooldown + 0.010) <
	                                  analTable[min].cooldown) then
		    -- the 10 ms slop is here to catch the case where
		    -- a timer tick occurs while the analysis add
		    -- routines were running, which will cause us to
		    -- flip between spells like we cannot make up our
		    -- minds.  The tick appears to be about 1ms (a
		    -- little over), so 10ms should be more than
		    -- plenty, and still be fast enough not to matter
		    -- to the user.
		    min = i;
		end
	    end
        end

	-- if min is 0, there is nothing they can cast
	if (min == 0) then
	    if DestroDummies.debug then
		print("Nothing is castable");
	    end
	    return nil, 0, nil;
	end

	-- Is this an item?  If so, then it must be castable
	if analTable[min].is_item then
	   if DestroDummies.debug then
	      print("Selected item ", analTable[min].spell)
	   end
	   return analTable[min].spell, analTable[min].cooldown, analTable[min].is_item;
	end

	-- Can they cast this spell?
	local usable, no_mana = IsUsableSpell(analTable[min].spell);

	-- note that we used to look for spells that eat health found
	-- by (ptype == -2) apparently, and checked
	-- UnitHealth("player") to see if we could cast it.  Now that
	-- we just call IsUsableSpell(), I have removed that logic
	--
	if usable then
	    -- they can cast this.  return it
	    if DestroDummies.debug then
		print("Selected ", analTable[min].name, " (#",
		      analTable[min].spell, ")");
	    end
	    return analTable[min].spell, analTable[min].cooldown, analTable[min].is_item;
	end

	-- Not usable.  If because of mana starvation, return the alternate
	if no_mana then
	    -- if they have a power alt spell off cooldown, return that one
	    if (powerAlt ~= nil and
	        DestroDummies:cooldownTime(powerAlt) == 0) then
		if DestroDummies.debug then
		    print("Selected the power alt spell: ", powerAlt);
		end
		return powerAlt, 0, nil
	    end

	    -- if they have a free spell off cooldown, return that one
	    if (freeSpell ~= nil and
	        DestroDummies:cooldownTime(freeSpell) == 0) then
		if DestroDummies.debug then
		    print("Selected the free spell: ", freeSpell);
		end
		return freeSpell, 0, nil;
	    end

	    -- The do not have a castable powerAlt or free spell, so
	    -- return nothing while we wait for mana to rebuild
	    if DestroDummies.debug then
		print("Want to cast ",
		      DestroDummies.spell_map[analTable[min].spell],
		      " (#",analTable[min].spell,
		      "), but we dinna have enough power.");
	    end
	    return nil, 0, nil;
	end

	-- Not usable for reasons other than mana.  Kill the entry and
	-- find the next best.  This should never happen, as we
	-- checked in analysisPick() before adding it
	if (DestroDummies.debug) then
	    print("    spell ", analTable[min].name, " not known????");
	end
	analTable[min].killed = true;
    end
end


-- Return the global cooldown start time and duration
--
-- Currently, we use a hack.  It appears that "Shoot" (spell 5019) is
-- on the global cooldown, and that GetSpellCooldown() will return the
-- cooldown time even if the player does not know it.local function
-- GetGCD ()
function DestroDummies:GetGCD ()
    return GetSpellCooldown(5019);
end

-- return the amount of time before you can cast the passed spell (in seconds)
function DestroDummies:cooldownTime (spell)
    local now = GetTime();

    -- Convert times to seconds.  UnitCastingInfo() returns
    -- milliseconds, GetSpellCooldown() returns seconds.  Thank you
    -- Bliz.  (We assume that if one is nil, the other must be, or
    -- it is a bug in the API)
    local curspell,_,_,_,curStartTime,curEndTime,_ = UnitCastingInfo("player");
    if (curStartTime == nil or curEndTime == nil) then
	curStartTime = 0;
	curEndTime = 0;
    else
	curStartTime = curStartTime / 1000;
	curEndTime = curEndTime / 1000;
    end

    -- See when the currently-being-cast spell will be done; we cannot
    -- cast anything before then.  We test on curEndTime to
    -- avoid future Bliz bugs.
    --
    local cur_cooldown = 0;
    if (curspell ~= nil and now > curEndTime) then
	cur_cooldown = curEndTime - now;
    end

    -- compute the cooldown specific to this spell
    local start, duration, enabled = GetSpellCooldown(spell);
    local spell_cooldown = start + duration - now;
    if (spell_cooldown < 0) then	-- protect against Bliz bugs
	spell_cooldown = 0;
    end

    -- There are some spells, such as Druids Thrash, which are
    -- subject to the GCD, but GetSpellCooldown() does not know it.  If
    -- the spell is not known to be off the GCD, factor GCD into the
    -- spell cooldown.
    if not DestroDummies:spell_off_GCD(playerClass, spell) then
	local gcd_start, gcd_dur = DestroDummies:GetGCD();
	local gcd_cooldown = gcd_start + gcd_dur - now;
	if (spell_cooldown < gcd_cooldown) then
	    spell_cooldown = gcd_cooldown;
	end
    end

    -- the actual cooldown is the max of the two cooldowns
    if (spell_cooldown > cur_cooldown) then
	return spell_cooldown;
    else
	return cur_cooldown;
    end
end


-- return the amount of power we will have at the end of the current cast
--[[
function DestroDummies:unitPowerAfterCast ()
   local power = UnitPower("player");

   -- What are we currently casting
   local name,_,_,_,_,endTime,isTrade = UnitCastingInfo("player");
   if (name and not isTrade) then
      -- How much more power will we have when we are done casting?
      local delta = (endTime / 1000) - GetTime();
      if (delta > 0) then
	 local regenNorm, regenCombat = GetPowerRegen();
	 local regen;
	 if (UnitAffectingCombat("player")) then
	    regen = regenCombat;
	 else
	    regen = regenNorm;
	 end
	 power = power + (regen * delta);
      end
      if power > UnitPowerMax("player") then
	 power = UnitPowerMax("player");
      end

      -- How much power will this take?
      local newName,_,_,cost = GetSpellInfo(name);
      if newName then	-- this should always be true
	 power = power - cost;
      end
   end
   return power;
end
--]]


/*
--  Return a guess at the distance to the target, along with the min and max
--  possible distances
--
--  This is made considerably harder than it needs to be by Bliz preventing
--  GetPlayerMapPosition from working with the "target" unit.  We have to
--  rely on rough guesses
--
--  CheckInteractDistance
--	1: 28 yds
-- 	2: 11.1 yds (actually, 11.11, but we are *so* not that accurate)
--	3: 9.9 yds
--
--  We also have some spells with known ranges we can use to test
--
--  We will end up with a (hopefully narrow) range of distances the mob can
--  be at.  If the distance between us and the mob's target is in that range,
--  assume the mob is at exactly that distance (ie., it is on the tank).
--  Otherwise return the average.
*/
function DestroDummies:distanceToTarget()
    local minrange = 0;
    local maxrange = 100;

    if (CheckInteractDistance("target",3)) then
	maxrange = 9.9;
    else
	minrange = 10;
	if (CheckInteractDistance("target",2)) then
	    maxrange = 11.1;
	else
	    minrange = 11.2;
	    if (CheckInteractDistance("target",1)) then
		maxrange = 28;
	    else
		minrange = 28.1;
	    end
	end
    end

    -- Now walk our Ranged spells to see which can help fix the range
    local sn;
    for sn = 1,rangeSpells.count do
	-- for easier-to-read code
	local smin = rangeSpells[sn].min;
	local smax = rangeSpells[sn].max;

	if (smin > minrange or (smax > 0 and smax < maxrange)) then
	    local rng = IsSpellInRange(rangeSpells[sn].spell, "target");
	    -- range of nil is invalid range
	    if (rng ~= nil) then
		if (rng == 0) then
		    -- out of range.  If our window overlaps just one end
		    -- of the range, we can shift that end.  If it overlaps
		    -- both, then we do not actually know which side is off.
		    local smin_inrange, smax_inrange;
		    smin_inrange = ((smin >= minrange) and (smin <= maxrange));
		    smax_inrange = ((smax >= minrange) and (smax <= maxrange));

		    if (not smin_inrange and not smax_inrange) then
		       --// we already knew this spell wouldn't work.  we
		       --// shouldn't have bothered trying
		    elseif (smin_inrange and smax_inrange) then
		       --// they could either be too close or too far.  maybe
		       --// we should come up with a way to come back and try
		       --// this again when the possible range has shrunk?
		    elseif (smin_inrange) then
		       --// We know that the spell minimum was within our
		       --// range, but the spell max was not.  That means that
		       --// we could not be too far away (the maximum is beyond
		       --// our possible distance).  So the problem must be
		       --// that we are too close).  That means that the max
		       --// distance we can be is less than the current spells'
		       --// minimum.  This sounds wrong but it is actually 
		       --//true.
		       --
		       maxrange = smin - 0.1;
		    else
			-- the inverse of the above
			minrange = smax + 0.1;
		    end
		else
		    -- Spell is in range
		    if (smin > minrange) then
			minrange = smin;
		    end
		    if (smax < maxrange and smax > 0) then
			maxrange = smax;
		    end
		end
	    end
	end
    end

    return (minrange + maxrange) / 2, minrange, maxrange;
end


-- Some spells (like Fishing) have very useful ranges but cannot be used
-- for range checking, so they must be specifically excluded to keep them
-- from excluding other spells that would actually be useful
--
local ExcludeRangeSpells = {};
ExcludeRangeSpells[1] = "Fishing";
ExcludeRangeSpells.count = 0;

-- build our list of range-checking spells
function DestroDummies:initRangeCheckSpells()
    rangeSpells.count = 0;

    local i;
    for i = 1, MAX_SKILLLINE_TABS do
        local name, _, offset, numSpells = GetSpellTabInfo(i);

        if not name then
            break
        end

	-- for every spell on the tab
        for s = offset + 1, offset + numSpells do
            local spell, _ = GetSpellBookItemName(s, BOOKTYPE_SPELL);
	    if (spell == nil) then
		-- should never happen
		break;
	    end

	    local sname,_,_,_,_,_,_,min,max = GetSpellInfo(spell);
	    if (spell ~= sname) then
		-- should never happen
		break;
	    end

	    local xno;
	    local xclude_this = false;
	    for xno = 1,ExcludeRangeSpells.count do
		if (ExcludeRangeSpells[xno] == spell) then
		    xclude_this = true;
		end
	    end

	    if not xclude_this and min ~= nil and max ~= nil then
		local useful = true;
	    	local spellNo;
	    	for spellNo = 1,rangeSpells.count do
		    if (rangeSpells[spellNo].spell == spell) then
			useful = false;
			break;
		    end
		end
		if useful then
		    rangeSpells.count = rangeSpells.count + 1;
		    rangeSpells[rangeSpells.count] = {};
		    rangeSpells[rangeSpells.count].spell = spell;
		    rangeSpells[rangeSpells.count].min = min;
		    rangeSpells[rangeSpells.count].max = max;
		end
	    end
        end
    end
end

-- find the spell in the spell book
local function get_spellid (spell)
    local tab, tabName, tabSpellOffset, tabNumSpells, spellIndex;

    for tab = 1, MAX_SKILLLINE_TABS do
	tabName, _, tabSpellOffset, tabNumSpells = GetSpellTabInfo(tab);
	if not tabName then
	    break
	end

	for spellIndex = tabSpellOffset + 1, tabSpellOffset + tabNumSpells do
	    if (type(spell) == "string") then
		-- match the name
		local name, subnm = GetSpellBookItemName(spellIndex,
							 BOOKTYPE_SPELL);
		if (name == spell) then
		    return spellIndex;
		end
	    else
		-- match the number
		local sType, id = GetSpellBookItemInfo(spellIndex,
						       BOOKTYPE_SPELL);
		if (id == spell) then
		    return spellIndex;
		end
	    end
	end
    end

    if (DestroDummies.debug) then
	print("get_spellid() of unknown spell ", spell);
    end
    return 0;
end


-- Sometimes we want to know the name or number for the spells, auras, etc.
-- This function allows you to get what is needed.
--
-- Parameters:
-- 	target:		name or number of thing
--	thing:		"spell", "aura", or "item"
--	format:		"string" or "number" (or "spellID" for spells)
--
function DestroDummies:munge (target, thing, format)
    -- simplify the chain of if/then/elseifs by pulling the error
    -- checking out to the top
    if (format ~= "string" and format ~= "number" and
	not (format == "spellID" and thing == "spell")) then
	-- this is an error
	if (DestroDummies.debug) then
	    print("Attempt to munge() to unknown format ", format);
	end
	return target;	-- better than nil, I suppose
    end

    -- it is simple if they are already in the requested format
    if (type(target) == format) then
	return target;
    end

    -- switch formats based on the thing type
    if (thing == "spell") then
	if (format == "spellID") then
	    -- look in the spell books for the id
	    return get_spellid(target);
	end
	return DestroDummies.spell_map[target];
    end

    if (thing == "aura") then
	return DestroDummies.aura_map[target];
    end

    if (thing == "item") then
	return DestroDummies.item_map[target];
    end

    -- This is an error
    if (DestroDummies.debug) then
	print("Attempt to munge() unknown thing: ", thing);
    end
    return target;	-- better than nil, I suppose
end
