/*
 *  Known strings (for spells, etc)
 *
 *  This is pretty clever, if I say so myself.  We are going to learn the
 *  local name for every spell by just knowing the mapping between the name
 *  in English and the spell id.  With this mapping, we can query the server
 *  to tell us the local name.
 */

#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include "builtin.h"
#include "strings.h"

// map between names and spell numbers
static struct spellInfo knownThings[] = {
    // things that are not spells--must be translated in the actual
    // Locale files
    { "Undead", -1 },
    { "Demon", -1 },
}, knownItems[] = {
    { "Oralius' Whispering Crystal",    118922 },
    { "Greater Draenic Agility Flask",  109153 },
    { "Greater Draenic Intellect Flask",109155 },
    { "Greater Draenic Strength Flask", 109156 },
    { "Hyper Augment Rune",             118630 },
    { "Focus Augment Rune",             118632 },
}, knownGlyphs[] = {
    { "Glyph of Exorcism",       41103 },
    { "Glyph of Ferocious Bite", 67600 },
    { "Glyph of Howling Blast",  45806 },
    { "Glyph of Life Tap",       63941 },
    { "Glyph of Maul",           54811 },
    { "Glyph of Thorns",         57862 },
    { "Glyph of Totem of Wrath", 63926 },
    { "Savagery",               171752 },       // Glyph of Savagery
}, knownTalents[] = {
    // Note that this mapping is not brought out into the Lua--it's
    // only used by the compiler.

    // Druid Talents
    { "Incarnation: King of the Jungle",        21705 },
    { "Soul of the Forest",      21702 },
    { "Bloodtalons",             21649 },
    { "Lunar Inspiration",	 21646 },

    // Hunter Talents
    { "Steady Focus",            19352 },
    { "Focusing Shot",           21662 },

    // Mage Talents
    { "Unstable Magic",          19300 },

    // Monk Talents
    { "Chi Brew",                19772 },
    { "Chi Explosion",           21190 },

    // Paladin Talents
    { "Empowered Seals",         21201 },
    { "Seraphim",                21202 },
    { "Final Verdict",           21672 },

    // Priest Talents
    { "Surge of Darkness",       21751 },
    { "Insanity",                21753 },

    // Rogue Talents
    { "Anticipation",            19250 },

    // Warlock Talents
    { "Demonbolt",               21694 },
    { "Cataclysm",               21181 },
    { "Blood Horror",            19290 },
    { "Soulburn: Haunt",         21180 },
    { "Charred Remains",         21695 },
}, knownAuras[] = {
    // from items
    { "Whispers of Insanity",   176151 },
    { "Greater Draenic Agility Flask",     156064 },
    //    { "Greater Draenic Intellect Flask",   ??? },
    //    { "Greater Draenic Strength Flask",    ??? },
    { "Draenor Versatility Flask",         156067 },
    { "Hyper Augmentation",     175456 },
    { "Focus Augmentation",     175457 },

    // from Death Knight spells
    { "Blood Charge",           114851 },
    { "Blood Plague",            55078 },
    { "Freezing Fog",            59052 },
    { "Frost Fever",             55095 },
    { "Killing Machine",         51124 },
    { "Shadow Infusion",         91342 },
    { "Sudden Doom",             81340 },

    // from Druid spells, and native Druid auras
    { "Clearcasting",           135700 },       // omen of clarity proc
    { "Predatory Swiftness",     69369 },
    { "Moonfire",               155625 },
    { "Rake",                   155722 },
    { "Pulverize",              158792 },
    { "Thrash-Bear",             77758 },	// goddamn fucking Bliz has
    { "Thrash-Kitty",           106830 },	// different numbers for Thrash

    // from Hunter spells
    { "Steady Focus",           177668 },

    // from Mage spells
    { "Frost Bomb",             112948 },

    // from Monk spells
    { "Combo Breaker: Blackout Kick",   116768 },
    { "Combo Breaker: Tiger Palm",      118864 },
    { "Combo Breaker: Chi Explosion",   159407 },       // from wowhead
    { "Tiger Power",            125359 },
    { "Tigereye Brew",          125195 },

    // from Rogue Spells
    { "Blindside",              121153 },
    { "Leaching Poison",        108211 },
    { "Crippling Poison",         3408 },
    { "Wound Poison",             8769 },
    { "Deadly Poison",            2823 },
    { "Stealth",                  1784 },
    { "Hemorrhage",              16511 },

    // from Warlock spells
    { "Corruption",             146739 },
    { "Molten Core",            122355 },
    { "Haunting Spirits",       157698 },       // from wowdb
    { "Dark Soul: Misery",      113860 },
    { "Immolate",               157736 },
}, knownSpells[] = {
    // Spells from spellbooks.  Note that the DestroDummies:showMySpells()
    // debugging function is by far the best way to gather this information.
    // The spell numbers on Wowhead cannot be relied on

    // Common spells
    { "Auto Attack",		  6603 },
    { "Shoot",                    5019 },

    // Death Knight Spells
    { "Anti-Magic Shell",        48707 },
    { "Army of the Dead",        42650 },
    { "Blood Boil",              50842 },
    { "Blood Presence",          48263 },
    { "Blood Tap",               45529 },
    { "Bone Shield",             49222 },
    { "Dancing Rune Weapon",     49028 },
    { "Dark Command",            56222 },
    { "Dark Transformation",     63560 },
    { "Death and Decay",         43265 },
    { "Death Coil",              47541 },
    { "Death Grip",              49576 },
    { "Death Strike",            49998 },
    { "Defile",                 152280 },       // from wowhead
    { "Empower Rune Weapon",     47568 },
    { "Festering Strike",        85948 },
    { "Frost Presence",          48266 },
    { "Frost Strike",            49143 },
    { "Horn of Winter",          57330 },
    { "Howling Blast",           49184 },
    { "Icebound Fortitude",      48792 },
    { "Icy Touch",               45477 },
    { "Obliterate",              49020 },
    { "Outbreak",                77575 },
    { "Plauge Leach",           123693 },
    { "Plague Strike",           45462 },
    { "Raise Dead",              46584 },
    { "Rune Tap",                48982 },
    { "Scourge Strike",          55090 },
    { "Soul Reaper",            130736 },
    { "Summon Gargoyle",         49206 },
    { "Unholy Blight",          115989 },
    { "Unholy Presence",         48265 },
    { "Crimson Scourge",         81136 },

    // Druid spells
    { "Bear Form",                5487 },
    { "Berserk",                106952 },
    { "Cat Form",                  768 },
    { "Cyclone",                 33786 },
    { "Dash",                     1850 },
    { "Entangling Roots",          339 },
    { "Faerie Fire",               770 },
    { "Ferocious Bite",          22568 },
    { "Frenzied Regeneration",   22842 },
    { "Growl",                    6795 },
    { "Healing Touch",            5185 },
    { "Hurricane",               16914 },
    { "Incarnation: King of the Jungle",        102543 },
    { "Maim",                    22570 },
    { "Mangle",                  33917 },
    { "Mark of the Wild",         1126 },
    { "Mighty Bash",            175683 },
    { "Moonfire",                 8921 },
    { "Prowl",                    5215 },
    { "Rake",                     1822 },
    { "Rebirth",                 20484 },
    { "Rejuvenation",              774 },
    { "Remove Corruption",        2782 },
    { "Revive",                  50769 },
    { "Rip",                      1079 },
    { "Savage Roar",             52610 },
    { "Shred",                    5221 },
    { "Skull Bash",             106839 },
    { "Soothe",                   2908 },
    { "Stampeding Roar",        106898 },
    { "Survival Instincts",      61336 },
    { "Swipe",                  106785 },
    { "Teleport: Moonglade",     18960 },
    { "Thrash",                 106832 },
    { "Tiger's Fury",             5217 },
    { "Travel Form",               783 },
    { "Typhoon",                175682 },
    { "Wrath",                    5176 },
    { "Faerie Swarm",           102354 },
    { "Dream of Cenarius",      158501 },
    { "Barkskin",                22812 },
    { "Lacerate",                33745 },
    { "Maul",                     6807 },
    { "Tooth and Claw",         135288 },
    { "Ursa Major",             159232 },
    { "Ysera's Gift",           145108 },
    { "Pulverize",               80313 },


    // Hunter Spells
    { "A Murder of Crows",      131894 },
    { "Arcane Shot",              3044 },
    { "Aspect of the Cheetah",    5118 },
    { "Aspect of the Fox",      172106 },
    { "Aspect of the Pack",      13159 },
    { "Wyvern Sting",            19386 },
    { "Mend Pet",                  136 },
    { "Revive Pet",                982 },
    { "Call Pet 1",                883 },
    { "Kill Command",            34026 },
    { "Rapid Fire",               3045 },
    { "Black Arrow",              3674 },
    { "Explosive Shot",          53301 },
    { "Stampede",               121818 },
    { "Thrill of the Hunt",     109306 },
    { "Glaive Toss",            117050 },
    { "Barrage",                120360 },
    { "Powershot",              109259 },
    { "Steady Shot",             56641 },
    { "Focusing Shot",          152245 },
    { "Cobra Shot",              77767 },
    { "Bestial Wrath",           19574 },
    { "Auto Shot",                  75 },
    { "Frenzy",                  19615 },
    { "Dire Beast",             120679 },
    { "Focus Fire",              82692 },
    { "Kill Shot",               53351 },
    { "Chimaera Shot",           53209 },
    { "Aimed Shot",              19434 },


    // Mage Spells
    { "Arcane Brilliance",        1459 },
    { "Blizzard",                   10 },
    { "Cone of Cold",              120 },
    { "Deep Freeze",             44572 },
    { "Frostbolt",                 116 },
    { "Frozen Orb",              84714 },
    { "Ice Barrier",             11426 },
    { "Ice Lance",                2136 },
    { "Icy Veins",               12472 },
    { "Cauterize",               86949 },
    { "Fingers of Frost",       112965 },
    { "Frost Armor",              7302 },
    { "Shatter",                 12982 },
    { "Amplify Magic",          159916 },
    { "Ice Shards",             165360 },
    { "Ice Ward",               175689 },
    { "Frostfire Bolt",          44614 },
    { "Brain Freeze",            44549 },
    { "Frost Bomb",             125430 },
    { "Summon Water Elemental",  31687 },
    { "Dalaran Brilliance",      61316 }, // from old numbers, not sure about
    { "Mirror Image",            55342 }, // also from old numbers

    // Monk spells
    { "Blackout Kick",          100784 },
    { "Chi Wave",               175693 },
    { "Detox",                  115450 },
    { "Fists of Fury",          113656 },
    { "Flying Serpent Kick",    101545 },
    { "Fortifying Brew",        115203 },
    { "Jab",                    100780 },
    { "Provoke",                115546 },
    { "Roll",                   109132 },
    { "Stance of the Fierce Tiger",     103985 },
    { "Surging Mist",           116694 },
    { "Tiger Palm",             100787 },
    { "Touch of Death",         115080 },
    { "Touch of Karma",         122470 },
    { "Tiger Strikes",          120272 },
    { "Tigereye Brew",          116740 },
    { "Nimble Brew",            137562 },
    { "Spear Hand Strike",      116705 },
    { "Energizing Brew",        115288 },
    { "Spinning Crane Kick",    101546 },
    { "Rising Sun Kick",        107428 },
    { "Expel Harm",             115072 },
    { "Chi Brew",               115399 },
    { "Zen Meditation",         115176 },
    { "Zen Sphere",             124081 },
    { "Chi Explosion",          152174 },       // from wowhead

    // Paladin spells
    { "Final Verdict",          157048 },
    { "Seal of Righteousness",  105361 },
    { "Seal of Command",        105361 },
    { "Seal of Insight",         20165 },
    { "Seal of Truth",           31801 },
    { "Seal of Justice",         20164 },
    { "Blessing of Might",       19740 },
    { "Templar's Verdict",       85256 },
    { "Exorcism",                  879 },
    { "Flash of Light",          19750 },
    { "Crusader Strike",         35395 },
    { "Divine Storm",            53385 },
    { "Empowered Divine Storm", 174718 },       // from wowhead
    { "Execution Sentence",     114157 },       // from wowhead
    { "Hammer of Wrath",         24275 },
    { "Judgment",                20271 },
    { "Seraphim",               152262 },       // from wowhead

    // Priest spells
    { "Devouring Plague",         2944 },
    { "Mind Blast",               8092 },
    { "Mind Flay",                 585 },
    { "Mind Spike",              73510 },
    { "Mindbender",              34433 },
    { "Power Word: Fortitude",   21562 },
    { "Power Word: Shield",         17 },
    { "Shadow Word: Death",      32379 },
    { "Shadow Word: Pain",         589 },
    { "Vampiric Touch",          34914 },
    { "Shadowform",              15473 },
    { "Shadowfiend",             34433 },
    { "Dispersion",              47585 },
    { "Vampiric Embrace",        15286 },
    { "Insanity",               139139 },
    { "Surge of Darkness",      162448 },


    // Rogue Spells
    { "Ambush",                   8676 },
    { "Envenom",                  2098 },
    { "Mutilate",                 1329 },
    { "Poisons",                    66 },
    { "Rupture",                  1943 },
    { "Slice and Dice",           5171 },
    { "Revealing Strike",        84617 },
    { "Eviscerate",               2098 },
    { "Adrenaline Rush",         13750 },
    { "Blade Flurry",            13877 },
    { "Preparation",             14185 },
    { "Backstab",                   53 },
    { "Shadow Dance",            51713 },
    { "Dispatch",                 1752 },
    { "Sinister Strike",          1752 },
    { "Hemorrhage",               1752 },


    // Warlock Spells
    { "Cataclysm",              152108 }, // these two come from my old spells,
    { "Demonbolt",              157695 }, // since my lock doesn't have them

    { "Blood Horror",           111397 },
    { "Banish",                    710 },
    { "Command Demon",          119898 },
    { "Corruption",                172 },
    { "Doom",                      603 },
    { "Create Healthstone",       6201 },
    { "Create Soulwell",         29893 },
    { "Dark Intent",            109773 },
    { "Dark Soul: Knowledge",   113861 },
    { "Dark Soul: Misery",      113860 },
    { "Dark Soul: Instability", 113858 },
    { "Demonic Leap",           109151 },
    { "Drain Life",                689 },
    { "Enslave Demon",            1098 },
    { "Fear",                     5782 },
    { "Hand of Gul'dan",        105174 },
    { "Health Funnel",             755 },
    { "Hellfire",                 1949 },
    { "Life Tap",                 1454 },
    { "Metamorphosis",          103958 },
    { "Sacrificial Pact",       108416 },
    { "Shadow Bolt",               686 },
    { "Shadowfury",             175707 },
    { "Soul Fire",                6353 },
    { "Soulshatter",             29858 },
    { "Soulstone",               20707 },
    { "Summon Abyssal",           1122 },
    { "Summon Demon",               10 },
    { "Summon Terrorguard",      18540 },
    { "Unending Breath",          5697 },
    { "Blood Pact",             166928 },
    { "Grimoire of Supremacy",  108499 },
    { "Molten Core",            122351 },
    { "Soul Harvest",           101976 },
    { "Soul Leech",             108370 },
    { "Wild Imps",              165392 },
    { "Chaos Wave",             105174 },
    { "Touch of Chaos",            686 },
    { "Summon Felguard",         30146 },
    { "Summon Wrathguard",      112870 },
    { "Haunt",                   48181 },
    { "Agony",                     980 },
    { "Unstable Affliction",     30108 },
    { "Summon Doomguard",        18540 },
    { "Soulburn",                74434 },
    { "Drain Soul",                686 },
    { "Summon Imp",                688 },
    { "Immolate",                  348 },
    { "Shadowburn",              17877 },
    { "Conflagrate",             17962 },
    { "Incinerate",              29722 },
    { "Chaos Bolt",             116858 },


    // Old spells...
    { "Battle Shout",            6673 },
    { "Roar of Courage",        93435 },
    { "Blessing of Kings",      20217 },
#if 0
    { "Aspect of the Fox",      172106 },
    { "Fingers of Frost",        44544 },
    { "Brain Freeze",            57761 },
    { "Hand of Gul'Dan",        105174 },
    { "Dark Intent",            109773 },
    { "Molten Core",            122351 },

    { "A Murder of Crows",      131894 },
    { "Abominable Might",       55972 },
    { "Adrenaline Rush",        13750 },
    { "Aimed Shot",             19434 },
    { "Ambush",                  8676 },
    { "Ancient Hysteria",       90355 },
    { "Arcane Barrage",         44425 },
    { "Arcane Blast",           10833 },
    { "Arcane Brilliance",       1459 },
    { "Arcane Missiles",         5143 },
    { "Arcane Power",           12042 },
    { "Arcane Shot",             3044 },
    { "Aspect of the Cheetah",   5118 },
    { "Aspect of the Iron Hawk",        109260 },
    { "Aspect of the Pack",     13159 },
    { "Auto Shot",                 75 },
    { "Backstab",                  53 },
    { "Bane of Agony",            980 },
    { "Bane of Doom",             603 },
    { "Bane of Havoc",          80240 },
    { "Barrage",                120360 },
    { "Bestial Wrath",          19574 },
    { "Black Arrow",             3674 },
    { "Blackout Kick",          100784 },
    { "Blade Flurry",           13877 },
    { "Blessing of Might",      19740 },
    { "Blood Charge",           114851 },
    { "Blood Presence",         48263 },
    { "Blood Shield",           77513 },
    { "Blood Tap",              45529 },
    { "Blood in the Water",     80863 },
    { "Bloodlust",               2825 },
    { "Bloodsurge",             46916 },
    { "Bone Shield",            49222 },
    { "Call Pet",                 883 },
    { "Cataclysm",              152108 },
    { "Chimaera Shot",          53209 },
    { "Chimera Shot",           53209 },
    { "Cobra Shot",             77767 },
    { "Combo Breaker",          115636 },
    { "Combustion",             11129 },
    { "Commanding Shout",         469 },
    { "Conflagrate",            17962 },
    { "Consecration",           26573 },
    { "Corruption",               172 },
    { "Crimson Scourge",        81136 },
    { "Crusader Aura",          32223 },
    { "Crusader Strike",        14517 },
    { "Dalaran Brilliance",     61316 },
    { "Dancing Rune Weapon",    49028 },
    { "Dark Command",           56222 },
    { "Dark Soul: Instability", 113858 },
    { "Dark Soul: Knowledge",   113861 },
    { "Dark Soul: Misery",      113860 },
    { "Dark Transformation",    93426 },
    { "Deadly Poison",          43581 },        // rogue debuff
    { "Death Coil",             47541 },
    { "Death Grip",             49576 },
    { "Death Strike",           49998 },
    { "Death Wish",             12292 },
    { "Death and Decay",        43265 },
    { "Deep Freeze",            44572 },
    { "Defile",                 152280 },
    { "Demon Soul",             77801 },
    { "Demonbolt",              157695 },
    { "Demoralizing Roar",         99 },
    { "Demoralizing Shout",      1160 },
    { "Devastate",              20243 },
    { "Devouring Plague",        2944 },
    { "Dire Beast",             120679 },
    { "Disengage",                781 },
    { "Divine Storm",           53385 },
    { "Doom",                   124913 },
    { "Drain Life",               689 },
    { "Earth Shock",             8042 },
    { "Empower Rune Weapon",    47568 },
    { "Energizing Brew",        115288 },
    { "Envenom",                32645 },
    { "Eviscerate",              2098 },
    { "Evocation",      12051 },
    { "Execute",        5308 },
    { "Execution Sentence",     114157 },
    { "Exorcism",       879 },
    { "Explosive Shot", 15495 },
    { "Feed Pet",       1539 },
    { "Fervor",                 82726 },
    { "Festering Strike",       85948 },
    { "Fire Blast",              2136 },
    { "Fire Elemental Totem",    2894 },
    { "Fire Nova",               1535 },
    { "Fireball",       133 },
    { "Fists of Fury",          117418 },
    { "Flame Shock",             8050 },
    { "Flying Serpent Kick",    123586 },
    { "Focus Fire",             82692 },
    { "Force of Nature",        6913 },
    { "Freeze",                 33395 },
    { "Freezing Fog", 59052 },
    { "Frenzy Effect",          19615 },
    { "Frost Armor",             7302 },
    { "Frost Bomb",             113092 },
    { "Frost Presence",         48266 },
    { "Frost Strike", 49143 },
    { "Frostbolt",      116 },
    { "Frostfire Bolt", 44614 },
    { "Frozen Orb",              84714 },
    { "Garrote",                  703 },
    { "Ghoul Summons",  45850 },
    { "Glaive Toss",            117050 },
    { "Gore",           35290 },
    { "Hammer of Wrath",        24275 },
    { "Haunt",  48181 },
    { "Health Funnel",  755 },
    { "Heating Up",     48107 },
    { "Hemorrhage",             16511 },
    { "Heroic Strike",  78 },
    { "Heroism",        23682 },
    { "Holy Wrath",     2812 },
    { "Howling Blast", 49184 },
    { "Hunter's Mark",  1130 },
    { "Ice Barrier",             11426 },
    { "Ice Lance",              30455 },
    { "Ice Ward",               111264 },
    { "Icy Touch", 45477 },
    { "Icy Veins",      12472 },
    { "Immolate",       348 },
    { "Incanter's Ward",        1463 },
    { "Incinerate",     19397 },
    { "Inferno Blast",  108853 },
    { "Jab",                    100780 },
    { "Judgement",              20271 },
    { "Kill Command",   34026 },
    { "Kill Shot",      53351 },
    { "Killing Machine", 51124 },
    { "Killing Spree",          51690 },
    { "Lacerate",       33745 },
    { "Lava Burst",     51505 },
    { "Lava Lash",      60103 },
    { "Life Tap",       1454 },
    { "Lightning Bolt", 403 },
    { "Lightning Shield",       324 },
    { "Living Bomb",    20475 },
    { "Mage Armor",     6117 },
    { "Magma Totem",    8187 },
    { "Maul",   6807 },
    { "Mend Pet",       136 },
    { "Metamorphosis",  36298 },
    { "Mind Blast",     8092 },
    { "Mind Flay",      15407 },
    { "Mirror Image",   55342 },
    { "Molten Armor",   30482 },
    { "Moonkin Form",   24858 },
    { "Multi-Shot",     2643 },
    { "Murder of Crows",        131894 },
    { "Mutilate",                1329 },
    { "Nature's Swiftness",     132158 },
    { "Nature's Vigil",         124974 },
    { "Nether Tempest", 114954 },
    { "Obliterate", 49020 },
    { "Omen of Clarity",        16864 },
    { "Outbreak",               77575 },
    { "Overkill",               58426 },
    { "Pestilence",             50842 },
    { "Pillar of Frost",        51271 },
    { "Plague Leech",           123693 },
    { "Plague Strike", 45462 },
    { "Power Word: Fortitude",  21562 },
    { "Power Word: Shield",     17 },
    { "Powershot",              109259 },
    { "Premeditation",          14183 },
    { "Preparation",            14185 },
    { "Presence of Mind",       12043 },
    { "Pulverize",              80313 },
    { "Pyroblast",      11366 },
    { "Raise Dead",     46584 },
    { "Rapid Fire",     3045 },
    { "Recklessness",            1719 },
    { "Recuperate",             73651 },
    { "Rend",   772 },
    { "Revealing Strike",       84617 },
    { "Revenge",        6572 },
    { "Revive Pet",     982 },
    { "Rupture",                 1943 },
    { "Savage Roar",    52610 },
    { "Scourge Strike",         55090 },
    { "Seal of Insight",        20165 },
    { "Seal of Justice",        20164 },
    { "Seal of Righteousness",  20154 },
    { "Seal of Truth",  31801 },
    { "Searing Totem",  3599 },
    { "Seed of Corruption",     27243 },
    { "Shadow Bolt",    686 },
    { "Shadow Dance",           51713 },
    { "Shadow Orb",             77487 },
    { "Shadow Trance",  17941 },
    { "Shadow Word: Death",     32379 },
    { "Shadow Word: Pain",      589 },
    { "Shadowburn",     17877 },
    { "Shadowfiend",            34433 },
    { "Shadowform",     15473 },
    { "Shamanistic Rage",       30823 },
    { "Shield Slam",    8242 },
    { "Shoot",                   5019 },
    { "Sinister Strike",         1752 },
    { "Slam",   1464 },
    { "Slice and Dice",          5171 },
    { "Solar Beam",             78675 },
    { "Soul Fire",      6353 },
    { "Soul Reaper",            130735 },
    { "Soulburn",               74434 },
    { "Stampede",               121818 },
    { "Starfall",               50288 },
    { "Starfire",       2912 },
    { "Starsurge",              78674 },
    { "Steady Focus",           177667 },
    { "Steady Shot",            56641 },
    { "Stealth",                 1784 },
    { "Stormstrike",    17364 },
    { "Summon Doomguard",       18540 },
    { "Summon Felguard",        30146 },
    { "Summon Felhunter",       691 },
    { "Summon Gargoyle",        49206 },
    { "Summon Imp",     688 },
    { "Summon Succubus",        712 },
    { "Summon Terrorguard",     112926 },
    { "Summon Water Elemental", 17162 },
    { "Sunfire",                93402 },
    { "Taunt",  355 },
    { "Templar's Verdict",      85256 },
    { "Thrill of the Hunt",     109306 },
    { "Thunder Clap",   6343 },
    { "Thunderclap",    8078 },
    { "Tiger Palm",             100787 },
    { "Tiger Power",            125359 },
    { "Tigereye Brew",          116740 },
    { "Time Warp",              80353 },
    { "Tooth and Claw",         135288 },
    { "Touch of Chaos",         156967 },
    { "Tricks of the Trade",    57934 },
    { "Trueshot Aura",  19506 },
    { "Unholy Presence",        48265 },
    { "Unleash Elements", 73680 },
    { "Unleash Flame",  73683 },
    { "Unstable Affliction",    30108 },
    { "Vampiric Embrace",       15286 },
    { "Vampiric Touch", 34914 },
    { "Vanish",                  1856 },
    { "Vendetta",               79140 },
    { "Victory Rush",   34428 },
    { "Water Shield",           52127 },
    { "Whirlwind",      1680 },
    { "Windfury Weapon",        8232 },
    { "Wyvern Sting",           19386 },
#endif
};

// Map between things and the auras they provide.  (Note that we don't
// need to map the cases where the thing names and the aura names are
// identical).
//
// This is only used by the compiler; it is not brought out to the Lua
// (nor should it be, to avoid internationalization issues)
static struct {
    char *thing_name;
    enum thingtype thing_type;
    char *aura;
} thing_aura_map[] = {
    { "Oralius' Whispering Crystal",    TH_ITEM,    "Whispers of Insanity" },
    { "Hyper Augment Rune",             TH_ITEM,    "Hyper Augmentation" },
    { "Focus Augment Rune",             TH_ITEM,    "Focus Augmentation" },
};


#define NELEM(x)        (sizeof(x) / sizeof(*(x)))

static int
compar (const void *p1, const void *p2)
{
    const struct spellInfo *sp1 = p1, *sp2 = p2;

    return strcasecmp(sp1->name, sp2->name);
}

static int
intcompar (const void *p1, const void *p2)
{
    const struct spellInfo *sp1 = p1, *sp2 = p2;

    return sp1->number - sp2->number;
}

int
isSpell (char *string)
{
    char stripped[1024];
    struct spellInfo target;

    // strip off leading and trailing quotes
    strcpy(stripped, &string[1]);
    stripped[strlen(stripped) - 1] = '\0';

    target.name = stripped;
    target.number = 0;

    // check spells
    if (bsearch(&target, knownSpells, NELEM(knownSpells),
                sizeof(*knownSpells), compar) != NULL)
        return 1;

    return 0;
}

int
isItem (char *string)
{
    char stripped[1024];
    struct spellInfo target;

    // strip off leading and trailing quotes
    strcpy(stripped, &string[1]);
    stripped[strlen(stripped) - 1] = '\0';

    target.name = stripped;
    target.number = 0;

    // check spells
    if (bsearch(&target, knownItems, NELEM(knownItems),
                sizeof(*knownItems), compar) != NULL) {
        return 1;
    }

    return 0;
}

// return the spell number, or 0 if unknown
int
thing_number (const enum thingtype type, const char * const name)
{
    char stripped[1024];
    struct spellInfo target, *result;
    struct spellInfo *ary;
    int nelem;

    switch (type) {
    case TH_SPELL:
        ary = knownSpells;
        nelem = NELEM(knownSpells);
        break;

    case TH_AURA:
        ary = knownAuras;
        nelem = NELEM(knownAuras);
        break;

    case TH_ITEM:
        ary = knownItems;
        nelem = NELEM(knownItems);
        break;

    case TH_GLYPH:
        ary = knownGlyphs;
        nelem = NELEM(knownGlyphs);
        break;

    case TH_TALENT:
        ary = knownTalents;
        nelem = NELEM(knownTalents);
        break;

    default:
        fprintf(stderr, "Unknown thing type %d\n", type);
        exit(1);
    }


    // strip off leading and trailing quotes if present
    if (*name == '"') {
        strcpy(stripped, &name[1]);
        stripped[strlen(name) - 2] = '\0';
    } else {
        strcpy(stripped, name);
    }

    target.name = stripped;
    target.number = 0;

    if ((result = bsearch(&target, ary, nelem, sizeof(*ary), compar)) != NULL){
        return result->number;
    }

    return 0;
}

// return the talent number for the passed-in thing (or 0 if unknown)
int
make_talent (const struct thing * const thing)
{
    int i;

    // get the easy case out of the way first
    if (thing->type == TH_TALENT)
        return thing->index;

    // Look for a talent of the same name
    for (i = 0 ; i < NELEM(knownTalents) ; i++) {
        if (!strcmp(knownTalents[i].name, thing->name))
            return knownTalents[i].number;
    }

    // Return 0 to show we didn't find it
    return 0;
}

// return the aura number for the passed-in thing
static int
make_aura_real (const struct thing * const thing)
{
    int i;

    // get the easy case out of the way first
    if (thing->type == TH_AURA)
        return thing->index;

    // do we have a specific mapping for this thing?
    for (i = 0 ; i < NELEM(thing_aura_map) ; i++) {
        if (thing_aura_map[i].thing_type == thing->type &&
            !strcmp(thing_aura_map[i].thing_name, thing->name)) {
            // yes, return the matching aura
            return thing_number(TH_AURA, thing_aura_map[i].aura);
        }
    }

    // Look for an aura of the same name
    for (i = 0 ; i < NELEM(knownAuras) ; i++) {
        if (!strcmp(knownAuras[i].name, thing->name))
            return knownAuras[i].number;
    }

    // Return the thing's number
    return thing->index;
}

int
make_aura (const struct thing * const thing)
{
    int result;

    // make_aura_real() should never return 0, yet it does right now.  I'll
    // fix that bug, but it's still worth wrapping it to catch future bugs.
    if ((result = make_aura_real(thing)) == 0) {
        fprintf(stderr, "%s:%d: returning aura 0!\n", filename, lineno);
        had_error = true;
    }
    return result;
}

static struct dup_info {
    char *name;
    int number;
    struct dup_info *next;
} *first_dup = NULL;

static void
have_dup_spell (struct spellInfo *spell)
{
    struct dup_info *cur;

    // do we already know about this?
    for (cur = first_dup ; cur != NULL ; cur = cur->next) {
        if (cur->number == spell->number && !strcmp(cur->name, spell->name))
            return;     // known
    }
    if ((cur = calloc(1, sizeof *cur)) == NULL) {
        perror("calloc");
        exit(1);
    }
    if ((cur->name = strdup(spell->name)) == NULL) {
        perror("strdup");
        exit(1);
    }
    cur->number = spell->number;

    // put at head of list
    cur->next = first_dup;
    first_dup = cur;
}


void
verifyStrings (void)
{
    bool trouble = false;
    int i;

#define SORT_N_CHECK(ary)       {                                       \
        qsort((ary), NELEM(ary), sizeof(*(ary)), intcompar);            \
        for (i = 1 ; i < NELEM(ary) ; i++) {                            \
            if ((ary)[i - 1].number == (ary)[i].number) {               \
                have_dup_spell(&(ary)[i - 1]);                          \
                have_dup_spell(&(ary)[i]);                              \
            }                                                           \
        }                                                               \
        qsort((ary), NELEM(ary), sizeof(*(ary)), compar);               \
        for (i = 1 ; i < NELEM(ary) ; i++) {                            \
            if (!strcasecmp((ary)[i - 1].name, (ary)[i].name)) {        \
                fprintf(stderr, "Duplicate entries for %s\n", (ary)[i].name); \
                trouble = true;                                         \
            }                                                           \
        }                                                               \
    }


    SORT_N_CHECK(knownSpells);
    SORT_N_CHECK(knownItems);
    SORT_N_CHECK(knownThings);
    SORT_N_CHECK(knownAuras);
    SORT_N_CHECK(knownGlyphs);
    SORT_N_CHECK(knownTalents);

    if (trouble)
        exit(1);
}

void
genLocale (void)
{
    int i;

    printf("-- Warning: this file is autogenerated.  All changes made here\n");
    printf("-- will be lost\n\n");
    printf("local spellIndex = 0\n");
    printf("local itemIndex = 0\n");
    printf("local English2SpellIndex = {};\n");
    printf("local English2ItemIndex = {};\n");
    printf("local knownSpellIndex = {};\n\n");
    printf("local knownItemIndex = {};\n\n");

    // this is the new mechanism--I think we should be able to get rid of
    // the English and known stuff
    printf("DestroDummies.spell_map = {};\n");
    printf("DestroDummies.item_map = {};\n");
    printf("DestroDummies.aura_map = {};\n");
    printf("local local_name;\n");      // name in local language

    for (i = 0 ; i < NELEM(knownSpells) ; i++) {
        if (knownSpells[i].number != -1) {
            printf("English2SpellIndex[\"%s\"] = %d;\n",
                   knownSpells[i].name, knownSpells[i].number);
        }
        printf("local_name = GetSpellInfo(%d);\n", knownSpells[i].number);
        printf("if local_name then\n");
        printf("    DestroDummies.spell_map[local_name] = %d;\n",
               knownSpells[i].number);
        printf("    DestroDummies.spell_map[%d] = local_name;\n",
               knownSpells[i].number);
        printf("else\n");
        printf("    DestroDummies.spell_map[%d] = \"%s\";\n",
               knownSpells[i].number, knownSpells[i].name);
        printf("    print(\"Destruction For Dummies: Unknown spell #%d\");\n",
               knownSpells[i].number);
        printf("end\n");
        printf("DestroDummies.spell_map[\"%s\"] = %d;\n",
               knownSpells[i].name, knownSpells[i].number);
    }

    for (i = 0 ; i < NELEM(knownAuras) ; i++) {
        printf("DestroDummies.aura_map[\"%s\"] = %d;\n",
                   knownAuras[i].name, knownAuras[i].number);
        printf("DestroDummies.aura_map[%d] = \"%s\";\n",
               knownAuras[i].number, knownAuras[i].name);
    }

    for (i = 0 ; i < NELEM(knownItems) ; i++) {
        if (knownItems[i].number != -1) {
            printf("English2ItemIndex[\"%s\"] = %d;\n",
                   knownItems[i].name, knownItems[i].number);
        }
        printf("local_name = GetItemInfo(%d);\n", knownItems[i].number);
        printf("if local_name then\n");
        printf("    DestroDummies.item_map[local_name] = %d;\n",
               knownItems[i].number);
        printf("    DestroDummies.item_map[%d] = local_name;\n",
               knownItems[i].number);
        printf("else\n");
        printf("    DestroDummies.item_map[%d] = \"%s\";\n",
               knownItems[i].number, knownItems[i].name);
        printf("    print(\"Destruction For Dummies: Unknown item #%d\");\n",
               knownItems[i].number);
        printf("end\n");
        printf("DestroDummies.item_map[\"%s\"] = %d;\n",
               knownItems[i].name, knownItems[i].number);
    }

    for (i = 0 ; i < NELEM(knownThings) ; i++) {
        if (knownThings[i].number != -1)
            printf("English2ThingIndex[\"%s\"] = %d;\n",
                   knownThings[i].name, knownThings[i].number);
    }

    printf("\n");
    for (i = 0 ; i < NELEM(knownSpells) ; i++) {
        printf("knownSpellIndex[spellIndex] = \"%s\";\tspellIndex = spellIndex + 1;\n",
               knownSpells[i].name);
    }
    for (i = 0 ; i < NELEM(knownItems) ; i++) {
        printf("knownItemIndex[itemIndex] = \"%s\";\titemIndex = itemIndex + 1;\n",
               knownItems[i].name);
    }


    printf("\n");
    printf("\n");
    printf("----------------------------------------------\n");
    printf("--\n");
    printf("--  Learn all the local names\n");
    printf("--\n");
    printf("\n");
    printf("for cur = 0, (spellIndex - 1) do\n");
    printf("    local curSpid, curSpell;\n");
    printf("    curSpell = knownSpellIndex[cur];\n");
    printf("    curSpid = English2SpellIndex[curSpell];\n");
    printf("\n");
    printf("    local name = GetSpellInfo(curSpid);\n");
    printf("    if name == nil then\n");
    printf("        print(\"Destruction for Dummies: cannot find spell info on \", curSpell);\n");
    printf("    else\n");
    printf("        DestroDummies.string[curSpell] = name;\n");
    printf("    end\n");
    printf("end\n");
    printf("\n");
    printf("for cur = 0, (itemIndex - 1) do\n");
    printf("    local curItemId, curItem;\n");
    printf("    curItem = knownItemIndex[cur];\n");
    printf("    curItemId = English2ItemIndex[curItem];\n");
    printf("\n");
    printf("    local name = GetItemInfo(curItemId);\n");
    printf("    if name == nil then\n");
    printf("        print(\"Destruction for Dummies: cannot find item info on \", curItem);\n");
    printf("    else\n");
    printf("        DestroDummies.string[curItem] = name;\n");
    printf("    end\n");
    printf("end\n");

    // learn about the different itemType and itemSubtype strings, as returned
    // by GetItemInfo()
    //
    // we do that by knowing one of each type of thing, and looking it up
    {
        struct {
            char *english;
            int exampleId;
        } stuff[] = {
            { "Daggers",                25101 },
            { "Fist Weapons",           17738 },
            { "One-Handed Axes",          811 },
            { "Two-Handed Axes",          870 },
            { "One-Handed Swords",       8190 },
            { "Two-Handed Swords",       2084 },
            { "One-Handed Maces",       36694 },
            { "Two-Handed Maces",       25135 },
            { "Staves",                 50040 },
            { "Guns",                   25639 },
            { "Bows",                   28772 },
            { "Crossbows",              28504 },
            { "Thrown",                 27916 },
            { "Wands",                  28673 },
        };
        int i;

        printf("\n\n");
        printf("-- Get the localized names for weapons\n");
        printf("local itemType, itemSubtype;\n");
        printf("_,_,_,_,_, itemType = GetItemInfo(%d);\n",
               stuff[0].exampleId);
        printf("DestroDummies.string[\"Weapon\"] = itemType;\n");
        for (i = 0 ; i < NELEM(stuff) ; i++) {
            printf("_,_,_,_,_,_, itemSubtype = GetItemInfo(%d);\n",
                   stuff[i].exampleId);
            printf("DestroDummies.string[\"%s\"] = itemSubtype;\n",
                   stuff[i].english);
        }
    }

    printf("\n");
    printf("\n");
    printf("-- Translate between the English strings we use internally and the localized \n");
    printf("-- strings.  Return the English for strings for which we do not have\n");
    printf("-- translations.\n");
    printf("function DestroDummies:xlateString (string)\n");
    printf("    if DestroDummies.string[string] == nil then\n");
    printf("        -- We are missing a translation.  Just return the English\n");
    printf("        return string;\n");
    printf("    end\n");
    printf("\n");
    printf("    return DestroDummies.string[string];\n");
    printf("end\n");

    printf("\n\n");
    printf("-- A debugging routine to find a spell number, given the name\n");
    printf("function DestroDummies:findSpellNo (string)\n");
    printf("    local cur, max, spell;\n");
    printf("\n");
    printf("    max = 200000            -- the highest number to try\n");
    printf("    for cur = 0, max do\n");
    printf("        spell = GetSpellInfo(cur);\n");
    printf("    if spell == string then\n");
    printf("        print(\"Spell \", spell, \" is number \", cur);\n");
    printf("        return;\n");
    printf("            end\n");
    printf("    end\n");
    printf("    print(\"Spell \", string, \" is unknown\");\n");
    printf("end\n");

    printf("\n\n");
    printf("-- A debugging routine to verify that all the spell numbers are correct\n");
    printf("-- This is never called--it is provided as a test mechanism for the author\n");
    printf("function DestroDummies:verifyCodes ()\n");
    printf("    local cur, curSpid, curSpell;\n");
    printf("    for cur = 0, (spellIndex - 1) do\n");
    printf("        curSpell = knownSpellIndex[cur];\n");
    printf("        curSpid = English2SpellIndex[curSpell];\n");
    printf("\n");
    printf("        local name = GetSpellInfo(curSpid);\n");
    printf("        if name ~= curSpell then\n");
    printf("            print(\"Destruction for Dummies: incorrect spell number \", curSpid, \": my name: \", curSpell, \", WoW Name: \", name);\n");
    printf("        end\n");
    printf("    end\n");
    printf("end\n");

    printf("\n");


    // maybe these are unique after all?
#if 0
    // Now generate the function that gets called whenever talents are
    // changed, to uniquify the spell table.  Sigh.
    printf("function DestroDummies:uniquify_spells()\n");
    printf("    local id;\n");

    struct dup_info *cur;
    for (cur = first_dup ; cur != NULL ; cur = cur->next) {

    local id;
    if (DestroDummies:utilTalent("Glaive Toss")) then
        id = DestroDummies.spell_map["Glaive Toss"];
        DestroDummies.spell_map[id] = "Glaive Toss";
    end
    if (DestroDummies:utilTalent("Barrage")) then
        id = DestroDummies.spell_map["Barrage"];
        DestroDummies.spell_map[id] = "Barrage";
    end
    if (DestroDummies:utilTalent("Powershot")) then
        id = DestroDummies.spell_map["Powershot"];
        DestroDummies.spell_map[id] = "Powershot";
    end

end
#endif

}
